# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import sys

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import cupy as cp  # noqa: F401
import awkward as ak  # noqa: F401


def test_cupy_refcount():
    o = cp.arange(10)
    i = ak.layout.Index64.from_cupy(o)
    assert sys.getrefcount(o) == 3
    i2 = ak.layout.Index64.from_cupy(o)
    assert sys.getrefcount(o) == 4
    i3 = ak.layout.Index64.from_cupy(o)
    assert sys.getrefcount(o) == 5
    i4 = ak.layout.Index64.from_cupy(o)
    assert sys.getrefcount(o) == 6
    i5 = ak.layout.Index64.from_cupy(o)
    assert sys.getrefcount(o) == 7
    i6 = ak.layout.Index64.from_cupy(o)
    assert sys.getrefcount(o) == 8
    i7 = ak.layout.Index64.from_cupy(o)
    assert sys.getrefcount(o) == 9
    i8 = ak.layout.Index64.from_cupy(o)
    assert sys.getrefcount(o) == 10
    i9 = ak.layout.Index64.from_cupy(o)
    assert sys.getrefcount(o) == 11

    del i9
    assert sys.getrefcount(o) == 10
    del i8
    assert sys.getrefcount(o) == 9
    del i7
    assert sys.getrefcount(o) == 8
    del i6
    assert sys.getrefcount(o) == 7
    del i5
    assert sys.getrefcount(o) == 6
    del i4
    assert sys.getrefcount(o) == 5
    del i3
    assert sys.getrefcount(o) == 4
    del i2
    assert sys.getrefcount(o) == 3
    del i
    assert sys.getrefcount(o) == 2



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    builder = ak.ArrayBuilder()
    with builder.list():
        with builder.record():
            builder.field("x").integer(1)
            builder.field("y").real(1.1)
        with builder.record():
            builder.field("x").integer(2)
            builder.field("y").real(2.2)
        with builder.record():
            builder.field("x").integer(3)
            builder.field("y").real(3.3)
    with builder.list():
        pass
    with builder.list():
        with builder.record():
            builder.field("x").integer(4)
            builder.field("y").real(4.4)
        with builder.record():
            builder.field("x").integer(5)
            builder.field("y").real(5.5)
            builder.field("z").string("five")
    with builder.list():
        with builder.record():
            builder.field("x").integer(6)
            builder.field("y").real(6.6)
            builder.field("z").string("six")
    with builder.list():
        with builder.record():
            builder.field("x").integer(7)
            builder.field("y").real(7.7)
        with builder.record():
            builder.field("x").integer(8)
            builder.field("y").real(8.8)
    assert ak.to_list(builder) == [
        [
            {"x": 1, "y": 1.1, "z": None},
            {"x": 2, "y": 2.2, "z": None},
            {"x": 3, "y": 3.3, "z": None},
        ],
        [],
        [{"x": 4, "y": 4.4, "z": None}, {"x": 5, "y": 5.5, "z": "five"}],
        [{"x": 6, "y": 6.6, "z": "six"}],
        [{"x": 7, "y": 7.7, "z": None}, {"x": 8, "y": 8.8, "z": None}],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    i = ak.Array([0, 1, 2, 3, 4, 5])
    x = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
    assert x[i[3:4].mask[[True]]].tolist() == [4.4]
    assert x[i[3:4][[True]]].tolist() == [4.4]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


pytest.importorskip("pyarrow.parquet")


def test(tmp_path):
    one = ak.Array([[], [{"x": [{"y": 1}]}]])
    two = ak.Array([[{"x": []}, {"x": [{"y": 1}]}]])
    three = ak.Array([[{"x": [{"y": 1}]}], [], [{"x": [{"y": 2}]}]])

    ak.to_parquet(one, tmp_path / "one.parquet")
    ak.to_parquet(two, tmp_path / "two.parquet")
    ak.to_parquet(three, tmp_path / "three.parquet")

    lazy_one = ak.from_parquet(tmp_path / "one.parquet", lazy=True)
    lazy_two = ak.from_parquet(tmp_path / "two.parquet", lazy=True)
    lazy_three = ak.from_parquet(tmp_path / "three.parquet", lazy=True)

    assert lazy_one.tolist() == [[], [{"x": [{"y": 1}]}]]
    assert lazy_two.tolist() == [[{"x": []}, {"x": [{"y": 1}]}]]
    assert lazy_three.tolist() == [[{"x": [{"y": 1}]}], [], [{"x": [{"y": 2}]}]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

import awkward.forth


def test_user_defined_error():
    vm32 = awkward.forth.ForthMachine32(
        """
s" this is an error"
"""
    )

    assert (
        vm32.decompiled
        == """s" this is an error"
"""
    )


def test_user_defined_strings():
    vm32 = awkward.forth.ForthMachine32(
        """
s" this is a first string"
s" this is a second string"
s" this is a third string"
s" this is a forth string"
"""
    )

    assert vm32.string_at(0) == "this is a first string"
    assert vm32.string_at(1) == "this is a second string"
    assert vm32.string_at(2) == "this is a third string"
    assert vm32.string_at(3) == "this is a forth string"
    assert vm32.string_at(4) == "a string at 4 is undefined"
    assert vm32.string_at(-1) == "a string at -1 is undefined"


def test_user_defined_exception():
    vm32 = awkward.forth.ForthMachine32(
        """
variable x 0 x !
variable err

s" variable x reached its maximum 10"

: foo
    5 x +! x @
    dup 10 = if
        0 err ! err @ halt
    then
;

0
begin
    foo
    1+
again
"""
    )
    with pytest.raises(ValueError):
        vm32.run()
    assert vm32.stack[-1] == 0
    assert vm32.string_at(vm32.stack[-1]) == "variable x reached its maximum 10"


def test_undefined_error():
    vm32 = awkward.forth.ForthMachine32(
        """
variable x 0 x !

: foo
    5 x +! x @
    dup 10 = if
        halt
    then
;

0
begin
    foo
    1+
again
"""
    )
    with pytest.raises(ValueError) as err:
        vm32.run()
    assert (
        str(err.value)
        == "'user halt' in AwkwardForth runtime: user-defined error or stopping condition"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    form = {
        "class": "RecordArray",
        "contents": {
            "a": {
                "class": "ListOffsetArray64",
                "offsets": "i64",
                "content": {
                    "class": "NumpyArray",
                    "itemsize": 8,
                    "format": "l",
                    "primitive": "int64",
                    "form_key": "node2",
                },
                "form_key": "node0",
            },
            "b": {
                "class": "ListOffsetArray64",
                "offsets": "i64",
                "content": {
                    "class": "NumpyArray",
                    "itemsize": 8,
                    "format": "l",
                    "primitive": "int64",
                    "form_key": "node3",
                },
                "form_key": "node0",
            },
        },
        "form_key": "node1",
    }

    container = {
        "part0-node0-offsets": np.array([0, 2, 3, 3, 6], dtype=np.int64),
        "part0-node2-data": np.array([1, 2, 3, 4, 5, 6], dtype=np.int64),
        "part0-node3-data": np.array([10, 20, 30, 40, 50, 60], dtype=np.int64),
    }

    assert ak.from_buffers(form, 4, container).tolist() == [
        {"a": [1, 2], "b": [10, 20]},
        {"a": [3], "b": [30]},
        {"a": [], "b": []},
        {"a": [4, 5, 6], "b": [40, 50, 60]},
    ]

    assert ak.from_buffers(form, 4, container, lazy=True).tolist() == [
        {"a": [1, 2], "b": [10, 20]},
        {"a": [3], "b": [30]},
        {"a": [], "b": []},
        {"a": [4, 5, 6], "b": [40, 50, 60]},
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import json

# don't import awkward._connect._uproot in awkward/__init__.py!
import uproot

import awkward as ak


def can_optimize(interpretation, form):
    if isinstance(interpretation, uproot.interpretation.objects.AsObjects):
        jsonform = json.loads(form.tojson(verbose=True))
        if (
            jsonform["class"] == "ListOffsetArray64"
            and jsonform["parameters"].get("uproot")
            == {"as": "array", "header": True, "speedbump": False}
            and jsonform["content"]["class"] == "ListOffsetArray64"
            and jsonform["content"]["parameters"].get("uproot")
            == {"as": "vector", "header": False}
            and jsonform["content"]["content"]["class"] == "NumpyArray"
            and jsonform["content"]["content"]["inner_shape"] == []
            and (
                jsonform["content"]["content"].get("primitive") == "float64"
                or jsonform["content"]["content"].get("primitive") == "int32"
            )
        ):
            return True

    return False


def basket_array(form, data, byte_offsets, extra):
    import awkward._io

    # FIXME: uproot_issue_90 is just a placeholder, to show how it would be done

    return awkward._io.uproot_issue_90(
        form,
        ak.layout.NumpyArray(data),
        ak.layout.Index32(byte_offsets),
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import awkward as ak  # noqa: F401


def test_primitive_1():
    text = "int64"
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.PrimitiveType)
    assert (str(parsedtype)) == text


def test_primitive_2():
    text = 'int64[parameters={"wonky": ["parameter", 3.14]}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.PrimitiveType)
    assert (str(parsedtype)) == text


def test_unknown_1():
    text = "unknown"
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.UnknownType)
    assert (str(parsedtype)) == text


def test_unknown_2():
    text = 'unknown[parameters={"wonky": ["parameter", 3.14]}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.UnknownType)
    assert str(parsedtype) == text


def test_record_tuple_1():
    text = "(int64)"
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_tuple_2():
    text = '(int64[parameters={"wonky": ["bla", 1, 2]}])'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_tuple_3():
    text = '(int64, int64[parameters={"wonky": ["bla", 1, 2]}])'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_dict_1():
    text = '{"1": int64}'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_dict_2():
    text = '{"bla": int64[parameters={"wonky": ["bla", 1, 2]}]}'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_dict_3():
    text = '{"bla": int64[parameters={"wonky": ["bla", 1, 2]}], "foo": int64}'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_parmtuple_1():
    text = 'tuple[[int64[parameters={"xkcd": [11, 12, 13]}]], parameters={"wonky": ["bla", 1, 2]}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_parmtuple_2():
    text = 'tuple[[int64, int64], parameters={"wonky": ["bla", 1, 2]}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_struct_1():
    text = 'struct[["1"], [int64[parameters={"xkcd": [11, 12, 13]}]], parameters={"wonky": ["bla", 1, 2]}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_struct_2():
    text = 'struct[["1", "2"], [int64[parameters={"xkcd": [11, 12, 13]}], int64], parameters={"wonky": ["bla", 1, 2]}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_option_numpy_1():
    text = "?int64"
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_numpy_2():
    text = '?int64[parameters={"wonky": [1, 2, 3]}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_numpy_1_parm():
    text = 'option[int64, parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_numpy_2_parm():
    text = 'option[int64[parameters={"wonky": [1, 2]}], parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_unknown_1():
    text = "?unknown"
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_unknown_2():
    text = '?unknown[parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_unknown_1_parm():
    text = 'option[unknown, parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_unknown_2_parm():
    text = 'option[unknown, parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_regular_numpy_1():
    text = "5 * int64"
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RegularType)
    assert str(parsedtype) == text


def test_regular_numpy_2():
    text = '5 * int64[parameters={"bar": "foo"}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RegularType)
    assert str(parsedtype) == text


def test_regular_numpy_2_parm():
    text = '[0 * int64[parameters={"foo": "bar"}], parameters={"bla": "bloop"}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RegularType)
    assert str(parsedtype) == text


def test_regular_unknown_1_parm():
    text = '[0 * unknown, parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.RegularType)
    assert str(parsedtype) == text


def test_list_numpy_1():
    text = "var * float64"
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.ListType)
    assert str(parsedtype) == text


def test_list_numpy_1_parm():
    text = '[var * float64[parameters={"wonky": "boop"}], parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.ListType)
    assert str(parsedtype) == text


def test_union_numpy_empty_1():
    text = 'union[float64[parameters={"wonky": "boop"}], unknown]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.UnionType)
    assert str(parsedtype) == text


def test_union_numpy_empty_1_parm():
    text = 'union[float64[parameters={"wonky": "boop"}], unknown, parameters={"pratyush": "das"}]'
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.UnionType)
    assert str(parsedtype) == text


def test_arraytype_string():
    text = str(ak.Array([["one", "two", "three"], [], ["four", "five"]]).type)
    parsedtype = ak.types.from_datashape(text, True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_bytestring():
    text = str(ak.Array([[b"one", b"two", b"three"], [], [b"four", b"five"]]).type)
    parsedtype = ak.types.from_datashape(text, True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_categorical_1():
    text = str(
        ak.to_categorical(ak.Array(["one", "one", "two", "three", "one", "three"])).type
    )
    parsedtype = ak.types.from_datashape(text, True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_categorical_2():
    text = str(ak.to_categorical(ak.Array([1.1, 1.1, 2.2, 3.3, 1.1, 3.3])).type)
    parsedtype = ak.types.from_datashape(text, True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_record_1():
    text = str(
        ak.Array(
            [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
            with_name="Thingy",
        ).type
    )
    parsedtype = ak.types.from_datashape(text, True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_record_2():
    text = str(
        ak.Array(
            [[{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}], [], [{"x": 3, "y": 3.3}]],
            with_name="Thingy",
        ).type
    )
    parsedtype = ak.types.from_datashape(text, True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_1():
    text = str(ak.Array([[1, 2, 3], None, [4, 5]]).type)
    parsedtype = ak.types.from_datashape(text, True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_2():
    text = str(
        ak.with_parameter(ak.Array([[1, 2, 3], [], [4, 5]]), "wonky", "string").type
    )
    parsedtype = ak.types.from_datashape(text, True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_3():
    text = str(
        ak.with_parameter(
            ak.Array([[1, 2, 3], [], [4, 5]]), "wonky", {"other": "JSON"}
        ).type
    )
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_4():
    text = str(
        ak.with_parameter(ak.Array([[1, 2, 3], None, [4, 5]]), "wonky", "string").type
    )
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_5():
    text = str(
        ak.with_parameter(ak.Array([1, 2, 3, None, 4, 5]), "wonky", "string").type
    )
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_6():
    text = str(ak.with_parameter(ak.Array([1, 2, 3, 4, 5]), "wonky", "string").type)
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_7():
    text = str(ak.Array([1, 2, 3, None, 4, 5]).type)
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_8():
    text = str(
        ak.with_parameter(
            ak.Array([{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}]),
            "wonky",
            "string",
        ).type
    )
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_9():
    text = str(ak.Array([(1, 1.1), (2, 2.2), (3, 3.3)]).type)
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_10():
    text = str(
        ak.with_parameter(
            ak.Array([(1, 1.1), (2, 2.2), (3, 3.3)]), "wonky", "string"
        ).type
    )
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_11():
    text = str(ak.Array([[(1, 1.1), (2, 2.2)], [], [(3, 3.3)]]).type)
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_12():
    text = str(ak.to_regular(ak.Array([[1, 2], [3, 4], [5, 6]])).type)
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_13():
    text = str(
        ak.with_parameter(
            ak.to_regular(ak.Array([[1, 2], [3, 4], [5, 6]])), "wonky", "string"
        ).type
    )
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_14():
    text = str(
        ak.with_parameter(
            ak.Array([1, 2, 3, [1], [1, 2], [1, 2, 3]]), "wonky", "string"
        ).type
    )
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_15():
    text = str(
        ak.with_parameter(
            ak.Array([1, 2, 3, None, [1], [1, 2], [1, 2, 3]]), "wonky", "string"
        ).type
    )
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_16():
    text = str(ak.Array([1, 2, 3, None, [1], [1, 2], [1, 2, 3]]).type)
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_arraytype_17():
    text = str(ak.Array([1, 2, 3, None, [], [], []]).type)
    parsedtype = ak.types.from_datashape(text)
    assert str(parsedtype) == text


def test_string():
    text = "string"
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.ListType)
    assert str(parsedtype) == text


def test_hardcoded():
    text = "var * string"
    parsedtype = ak.types.from_datashape(text)
    assert isinstance(parsedtype, ak.types.ListType)
    assert str(parsedtype) == text


def test_record_highlevel():
    text = 'Thingy["x": int64, "y": float64]'
    parsedtype = ak.types.from_datashape(text, True)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    one = ak.Array([[{"x": 1}], [], [{"x": 2}]], with_name="One")
    two = ak.Array([[{"x": 1.1}], [], [{"x": 2.2}]], with_name="Two")
    assert (
        str(ak.with_name(ak.concatenate([one, two], axis=1), "All").type)
        == '3 * var * All["x": float64]'
    )
    assert (
        str(ak.with_name(ak.concatenate([one[1:], two[1:]], axis=1), "All").type)
        == '2 * var * All["x": float64]'
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array(
        ak.layout.ListOffsetArray64(
            ak.layout.Index64(np.r_[0, 4, 4, 6]),
            ak.layout.NumpyArray(np.arange(4 * 6, dtype=np.int64).reshape(6, 4)),
        )
    )
    ix = ak.Array([[[0], [1], [2], [3]], [], [[3], [2]]])
    assert array[ix].tolist() == [[[0], [5], [10], [15]], [], [[19], [22]]]
    assert str(array[ix].type) == "3 * var * var * int64"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_same_categories():
    categories = ak.Array(["one", "two", "three"])
    index1 = ak.layout.Index64(np.array([0, 2, 2, 1, 2, 0, 1, 0], dtype=np.int64))
    index2 = ak.layout.Index64(np.array([1, 1, 2, 1, 0, 0, 0, 1], dtype=np.int64))
    categorical1 = ak.layout.IndexedArray64(
        index1, categories.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.layout.IndexedArray64(
        index2, categories.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).tolist() == [
        False,
        False,
        True,
        True,
        False,
        True,
        False,
        False,
    ]


def test_different_categories():
    categories1 = ak.Array(["one", "two", "three"])
    categories2 = ak.Array(["three", "two", "one"])
    index1 = ak.layout.Index64(np.array([0, 2, 2, 1, 2, 0, 1, 0], dtype=np.int64))
    index2 = ak.layout.Index64(np.array([1, 1, 2, 1, 0, 0, 0, 1], dtype=np.int64))
    categorical1 = ak.layout.IndexedArray64(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.layout.IndexedArray64(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).tolist() == [
        False,
        False,
        False,
        True,
        True,
        False,
        False,
        False,
    ]


def test_one_extra():
    categories1 = ak.Array(["one", "two", "three", "four"])
    categories2 = ak.Array(["three", "two", "one"])
    index1 = ak.layout.Index64(np.array([0, 3, 0, 1, 2, 3, 1, 0], dtype=np.int64))
    index2 = ak.layout.Index64(np.array([1, 1, 2, 1, 0, 0, 0, 1], dtype=np.int64))
    categorical1 = ak.layout.IndexedArray64(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.layout.IndexedArray64(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).tolist() == [
        False,
        False,
        True,
        True,
        True,
        False,
        False,
        False,
    ]


def test_two_extra():
    categories1 = ak.Array(["one", "two", "three"])
    categories2 = ak.Array(["four", "three", "two", "one"])
    index1 = ak.layout.Index64(np.array([0, 2, 2, 1, 2, 0, 1, 0], dtype=np.int64))
    index2 = ak.layout.Index64(np.array([1, 3, 1, 1, 3, 3, 2, 1], dtype=np.int64))
    categorical1 = ak.layout.IndexedArray64(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.layout.IndexedArray64(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).tolist() == [
        False,
        False,
        True,
        False,
        False,
        True,
        True,
        False,
    ]


def test_option_same_categories():
    categories = ak.Array(["one", "two", "three"])
    index1 = ak.layout.Index64(np.array([0, 2, 2, 1, -1, 0, 1, 0], dtype=np.int64))
    index2 = ak.layout.Index64(np.array([1, 1, 2, 1, -1, -1, 0, 1], dtype=np.int64))
    categorical1 = ak.layout.IndexedOptionArray64(
        index1, categories.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.layout.IndexedOptionArray64(
        index2, categories.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).tolist() == [
        False,
        False,
        True,
        True,
        True,
        False,
        False,
        False,
    ]


def test_option_different_categories():
    categories1 = ak.Array(["one", "two", "three"])
    categories2 = ak.Array(["three", "two", "one"])
    index1 = ak.layout.Index64(np.array([0, 2, 2, 1, -1, -1, 1, 0], dtype=np.int64))
    index2 = ak.layout.Index64(np.array([1, 1, 2, 1, 0, -1, 0, 1], dtype=np.int64))
    categorical1 = ak.layout.IndexedOptionArray64(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.layout.IndexedOptionArray64(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).tolist() == [
        False,
        False,
        False,
        True,
        False,
        True,
        False,
        False,
    ]


def test_option_one_extra():
    categories1 = ak.Array(["one", "two", "three", "four"])
    categories2 = ak.Array(["three", "two", "one"])
    index1 = ak.layout.Index64(np.array([0, -1, 0, 1, 2, 3, 1, 0], dtype=np.int64))
    index2 = ak.layout.Index64(np.array([1, -1, -1, 1, 0, 0, 0, 1], dtype=np.int64))
    categorical1 = ak.layout.IndexedOptionArray64(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.layout.IndexedOptionArray64(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).tolist() == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
        False,
    ]


def test_option_two_extra():
    categories1 = ak.Array(["one", "two", "three"])
    categories2 = ak.Array(["four", "three", "two", "one"])
    index1 = ak.layout.Index64(np.array([0, -1, -1, 1, 2, 0, 1, 0], dtype=np.int64))
    index2 = ak.layout.Index64(np.array([1, -1, 1, 1, 3, 3, 2, 1], dtype=np.int64))
    categorical1 = ak.layout.IndexedOptionArray64(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.layout.IndexedOptionArray64(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).tolist() == [
        False,
        True,
        False,
        False,
        False,
        True,
        True,
        False,
    ]


def test_to_categorical_numbers():
    array = ak.Array([1.1, 2.2, 3.3, 1.1, 2.2, 3.3, 1.1, 2.2, 3.3])
    assert not ak.is_categorical(array)
    categorical = ak.to_categorical(array)
    assert ak.is_categorical(categorical)
    assert ak.to_list(array) == categorical.tolist()
    assert ak.to_list(categorical.layout.content) == [1.1, 2.2, 3.3]
    not_categorical = ak.from_categorical(categorical)
    assert not ak.is_categorical(not_categorical)
    assert ak.categories(categorical).tolist() == [1.1, 2.2, 3.3]


def test_to_categorical_nested():
    array = ak.Array([["one", "two", "three"], [], ["one", "two"], ["three"]])
    assert not ak.is_categorical(array)
    categorical = ak.to_categorical(array)
    assert ak.is_categorical(categorical)
    assert ak.to_list(array) == categorical.tolist()
    not_categorical = ak.from_categorical(categorical)
    assert not ak.is_categorical(not_categorical)
    assert ak.categories(categorical).tolist() == ["one", "two", "three"]


def test_to_categorical():
    array = ak.Array(
        ["one", "two", "three", "one", "two", "three", "one", "two", "three"]
    )
    assert not ak.is_categorical(array)
    categorical = ak.to_categorical(array)
    assert ak.is_categorical(categorical)
    assert ak.to_list(array) == categorical.tolist()
    assert ak.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.from_categorical(categorical)
    assert not ak.is_categorical(not_categorical)
    assert ak.categories(categorical).tolist() == ["one", "two", "three"]


def test_to_categorical_none():
    array = ak.Array(
        [
            "one",
            "two",
            "three",
            None,
            "one",
            "two",
            "three",
            None,
            "one",
            "two",
            "three",
            None,
        ]
    )
    assert not ak.is_categorical(array)
    categorical = ak.to_categorical(array)
    assert ak.is_categorical(categorical)
    assert ak.to_list(array) == categorical.tolist()
    assert ak.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.from_categorical(categorical)
    assert not ak.is_categorical(not_categorical)
    assert ak.categories(categorical).tolist() == ["one", "two", "three"]


def test_to_categorical_masked():
    content = ak.Array(
        [
            "one",
            "two",
            "three",
            "one",
            "one",
            "two",
            "three",
            "two",
            "one",
            "two",
            "three",
            "three",
        ]
    ).layout
    mask = ak.layout.Index8(
        np.array(
            [
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
            ]
        )
    )
    array = ak.Array(ak.layout.ByteMaskedArray(mask, content, valid_when=False))
    assert not ak.is_categorical(array)
    categorical = ak.to_categorical(array)
    assert ak.is_categorical(categorical)
    assert ak.to_list(array) == categorical.tolist()
    assert ak.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.from_categorical(categorical)
    assert not ak.is_categorical(not_categorical)
    assert ak.categories(categorical).tolist() == ["one", "two", "three"]


def test_to_categorical_masked_again():
    content = ak.Array(
        ["one", "two", "three", "one", "one", "two", "three", "two"]
    ).layout
    index = ak.layout.Index64(
        np.array([0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3], dtype=np.int64)
    )
    indexedarray = ak.layout.IndexedArray64(index, content)
    mask = ak.layout.Index8(
        np.array(
            [
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
            ]
        )
    )
    array = ak.Array(ak.layout.ByteMaskedArray(mask, indexedarray, valid_when=False))
    assert not ak.is_categorical(array)
    categorical = ak.to_categorical(array)
    assert ak.is_categorical(categorical)
    assert ak.to_list(array) == categorical.tolist()
    assert ak.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.from_categorical(categorical)
    assert not ak.is_categorical(not_categorical)
    assert ak.categories(categorical).tolist() == ["one", "two", "three"]


def test_typestr():
    if not ak._util.py27:
        assert (
            str(ak.type(ak.to_categorical(ak.Array([1.1, 2.2, 2.2, 3.3]))))
            == "4 * categorical[type=float64]"
        )
        assert (
            str(ak.type(ak.to_categorical(ak.Array([1.1, 2.2, None, 2.2, 3.3]))))
            == "5 * categorical[type=?float64]"
        )
        assert (
            str(ak.type(ak.to_categorical(ak.Array(["one", "two", "two", "three"]))))
            == "4 * categorical[type=string]"
        )
        assert (
            str(
                ak.type(
                    ak.to_categorical(ak.Array(["one", "two", None, "two", "three"]))
                )
            )
            == "5 * categorical[type=option[string]]"
        )


def test_zip():
    x = ak.Array([1.1, 2.2, 3.3])
    y = ak.Array(["one", "two", "three"])
    assert ak.zip({"x": x, "y": y}).tolist() == [
        {"x": 1.1, "y": "one"},
        {"x": 2.2, "y": "two"},
        {"x": 3.3, "y": "three"},
    ]
    y = ak.to_categorical(y)
    assert ak.zip({"x": x, "y": y}).tolist() == [
        {"x": 1.1, "y": "one"},
        {"x": 2.2, "y": "two"},
        {"x": 3.3, "y": "three"},
    ]


pyarrow = pytest.importorskip("pyarrow")


def test_arrow_nomask():
    array = ak.Array([1.1, 2.2, 3.3, 4.4, None])
    assert str(ak.type(ak.from_arrow(ak.to_arrow(array)))) == "5 * ?float64"
    assert str(ak.type(ak.from_arrow(ak.to_arrow(array[:-1])))) == "4 * ?float64"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.values_astype(ak.Array([1.1, 2.2, None, 3.3]), np.float32)
    assert str(ak.fill_none(array, np.float32(0)).type) == "4 * float32"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import os

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


pytest.importorskip("pyarrow.parquet")


def test_no_fields(tmp_path):
    one = ak.Array([[1, 2, 3], [], [4, 5]])
    two = ak.Array([[6], [7, 8, 9, 10]])

    ak.to_parquet(one, tmp_path / "file1.parquet")
    ak.to_parquet(two, tmp_path / "file2.parquet")
    assert not os.path.exists(tmp_path / "_common_metadata")
    assert not os.path.exists(tmp_path / "_metadata")

    no_metadata = ak.from_parquet(tmp_path)
    assert no_metadata.tolist() == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]

    no_metadata_lazy = ak.from_parquet(tmp_path, lazy=True)
    assert no_metadata_lazy.tolist() == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]

    ak.to_parquet.dataset(tmp_path)
    assert os.path.exists(tmp_path / "_common_metadata")
    assert os.path.exists(tmp_path / "_metadata")

    with_metadata = ak.from_parquet(tmp_path)
    assert with_metadata.tolist() == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]

    with_metadata_lazy = ak.from_parquet(tmp_path, lazy=True)
    assert with_metadata_lazy.tolist() == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]


def test_with_fields(tmp_path):
    one_list = [[{"x": 1}, {"x": 2}, {"x": 3}], [], [{"x": 4}, {"x": 5}]]
    two_list = [[{"x": 6}], [{"x": 7}, {"x": 8}, {"x": 9}, {"x": 10}]]
    one = ak.Array(one_list)
    two = ak.Array(two_list)

    ak.to_parquet(one, tmp_path / "file1.parquet")
    ak.to_parquet(two, tmp_path / "file2.parquet")
    assert not os.path.exists(tmp_path / "_common_metadata")
    assert not os.path.exists(tmp_path / "_metadata")

    no_metadata = ak.from_parquet(tmp_path)
    assert no_metadata.tolist() == one_list + two_list

    no_metadata_lazy = ak.from_parquet(tmp_path, lazy=True)
    assert no_metadata_lazy.tolist() == one_list + two_list

    ak.to_parquet.dataset(tmp_path)
    assert os.path.exists(tmp_path / "_common_metadata")
    assert os.path.exists(tmp_path / "_metadata")

    with_metadata = ak.from_parquet(tmp_path)
    assert with_metadata.tolist() == one_list + two_list

    with_metadata_lazy = ak.from_parquet(tmp_path, lazy=True)
    assert with_metadata_lazy.tolist() == one_list + two_list


pandas = pytest.importorskip("pandas")


def test_pandas(tmp_path):
    df = pandas.DataFrame(
        {"x": np.arange(10), "y": np.arange(10) % 5, "z": ["low"] * 5 + ["high"] * 5}
    )
    df.to_parquet(tmp_path, partition_cols=["z", "y"])

    a = ak.from_parquet(tmp_path)
    assert a.z.tolist() == ["high"] * 5 + ["low"] * 5  # alphabetical
    assert a.y.tolist() == ["0", "1", "2", "3", "4", "0", "1", "2", "3", "4"]
    assert a.x.tolist() == [5, 6, 7, 8, 9, 0, 1, 2, 3, 4]

    b = ak.from_parquet(tmp_path, lazy=True)
    assert b.z.tolist() == ["high"] * 5 + ["low"] * 5
    assert b.y.tolist() == ["0", "1", "2", "3", "4", "0", "1", "2", "3", "4"]
    assert b.x.tolist() == [5, 6, 7, 8, 9, 0, 1, 2, 3, 4]

    c = ak.from_parquet(tmp_path, include_partition_columns=False)
    assert ak.fields(c) == ["x"]

    d = ak.from_parquet(tmp_path, lazy=True, include_partition_columns=False)
    assert ak.fields(d) == ["x"]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_mine():
    array = ak.Array(
        ak.layout.ListOffsetArray64(
            ak.layout.Index64(np.array([1, 2, 4, 7], np.int64)),
            ak.layout.NumpyArray(
                np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
            ),
        )
    )
    assert (
        ak.argmax(array, axis=-1).tolist()
        == ak.argmax(ak.Array([[1.1], [2.2, 3.3], [4.4, 5.5, 6.6]]), axis=-1).tolist()
    )


def test_998():
    array = ak.Array(
        ak.layout.ListOffsetArray64(
            ak.layout.Index64(np.array([1, 2, 4, 7], np.int64)),
            ak.layout.NumpyArray(
                np.r_[
                    1.8125,
                    0.8125,
                    -0.9375,
                    1.1875,
                    -0.6875,
                    1.3125,
                    21.3125,
                ]
            ),
        )
    )

    assert ak.to_list(ak.argmax(array, axis=-1)) == [0, 1, 2]


def test_1000():
    array = ak.Array(
        ak.layout.ListOffsetArray64(
            ak.layout.Index64(np.array([1, 3, 5], np.int64)),
            ak.layout.ListOffsetArray64(
                ak.layout.Index64(np.array([0, 3, 5, 8, 10, 12], np.int64)),
                ak.layout.NumpyArray(
                    np.r_[
                        1.8125,
                        0.81252,
                        -0.937,
                        6.0,
                        -0.6875,
                        1.3125,
                        21.3125,
                        4.0,
                        9.8,
                        2.2,
                        33.0,
                        44.6,
                    ]
                ),
            ),
        )
    )

    assert ak.to_list(ak.argmax(array, axis=1)) == [[0, 1, 1], [1, 1]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array(
        ak.layout.RegularArray(
            ak.layout.NumpyArray(np.r_[1, 2, 3, 4, 5, 6, 7, 8, 9]), 3
        )
    )
    mask = ak.Array(
        ak.layout.NumpyArray(
            np.array([[True, True, True], [True, True, False], [True, False, True]])
        )
    )

    assert ak.mask(array, mask).to_list() == [[1, 2, 3], [4, 5, None], [7, None, 9]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


@pytest.mark.skipif(
    ak._util.py27 or ak._util.py35,
    reason="SOMETIMES doesn't work under Python 2.7 and 3.5",
)
def test():
    out = np.digitize(ak.Array([1.1, 2.2, 3.3, 4.4, 5.5]), ak.Array([0, 1, 3, 6]))
    assert isinstance(out, ak.Array)
    assert out.tolist() == [2, 2, 3, 3, 3]

    assert np.median(ak.Array([1.1, 2.2, 3.3, 4.4, 5.5])) == 3.3



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    assert ak.is_none(ak.Array([1, 2, 3, None, 5])).tolist() == [
        False,
        False,
        False,
        True,
        False,
    ]
    assert ak.is_none(ak.Array([[1, 2, 3], [], [None, 5]])).tolist() == [
        False,
        False,
        False,
    ]
    assert ak.is_none(ak.Array([[1, 2, 3], [], [None, 5]]), axis=1).tolist() == [
        [False, False, False],
        [],
        [True, False],
    ]
    assert ak.is_none(ak.Array([[1, None, 2, 3], [], [None, 5]]), axis=1).tolist() == [
        [False, True, False, False],
        [],
        [True, False],
    ]
    assert ak.is_none(ak.Array([[1, None, 2, 3], [], [None, 5]]), axis=-1).tolist() == [
        [False, True, False, False],
        [],
        [True, False],
    ]
    assert ak.is_none(ak.Array([[1, None, 2, 3], [], [None, 5]]), axis=-2).tolist() == [
        False,
        False,
        False,
    ]
    assert ak.is_none(
        ak.Array([[1, None, 2, 3], None, [None, 5]]), axis=-2
    ).tolist() == [False, True, False]
    with pytest.raises(ValueError):
        ak.is_none(ak.Array([[1, None, 2, 3], [], [None, 5]]), axis=-3)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.forms.numpyform import NumpyForm

np = ak.nplike.NumpyMetadata.instance()


class NumpyArray(Content):
    def __init__(self, data, identifier=None, parameters=None, nplike=None):
        self._nplike = ak.nplike.of(data) if nplike is None else nplike
        self._data = self._nplike.asarray(data)

        if (
            self._data.dtype not in ak._v2.types.numpytype._dtype_to_primitive
            and not issubclass(self._data.dtype.type, (np.datetime64, np.timedelta64))
        ):
            raise TypeError(
                "{0} 'data' dtype {1} is not supported; must be one of {2}".format(
                    type(self).__name__,
                    repr(self._data.dtype),
                    ", ".join(
                        repr(x) for x in ak._v2.types.numpytype._dtype_to_primitive
                    ),
                )
            )
        if len(self._data.shape) == 0:
            raise TypeError(
                "{0} 'data' must be an array, not {1}".format(
                    type(self).__name__, repr(data)
                )
            )

        self._init(identifier, parameters)

    @property
    def data(self):
        return self._data

    @property
    def shape(self):
        return self._data.shape

    @property
    def inner_shape(self):
        return self._data.shape[1:]

    @property
    def strides(self):
        return self._data.strides

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def nplike(self):
        return self._nplike

    @property
    def nonvirtual_nplike(self):
        return self._nplike

    @property
    def ptr(self):
        return self._data.ctypes.data

    Form = NumpyForm

    @property
    def form(self):
        return self.Form(
            ak._v2.types.numpytype._dtype_to_primitive[self._data.dtype],
            self._data.shape[1:],
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def typetracer(self):
        return NumpyArray(
            self.to(ak._v2._typetracer.TypeTracer.instance()),
            self._typetracer_identifier(),
            self._parameters,
            nplike=ak._v2._typetracer.TypeTracer.instance(),
        )

    def __len__(self):
        return len(self._data)

    def to(self, nplike):
        return nplike.asarray(self._data)

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<NumpyArray dtype="]
        out.append(repr(str(self.dtype)))
        if len(self._data.shape) == 1:
            out.append(" len=" + repr(str(self._data.shape[0])))
        else:
            out.append(
                " shape='({0})'".format(", ".join(str(x) for x in self._data.shape))
            )

        extra = self._repr_extra(indent + "    ")
        arraystr_lines = self._nplike.array_str(self._data, max_line_width=30).split(
            "\n"
        )
        if len(extra) != 0 or len(arraystr_lines) > 1:
            arraystr_lines = self._nplike.array_str(
                self._data, max_line_width=max(80 - len(indent) - 4, 40)
            ).split("\n")
            if len(arraystr_lines) > 5:
                arraystr_lines = arraystr_lines[:2] + [" ..."] + arraystr_lines[-2:]
            out.append(">")
            out.extend(extra)
            out.append("\n" + indent + "    ")
            out.append(("\n" + indent + "    ").join(arraystr_lines))
            out.append("\n" + indent + "</NumpyArray>")
        else:
            out.append(">")
            out.append(arraystr_lines[0])
            out.append("</NumpyArray>")

        out.append(post)
        return "".join(out)

    def toRegularArray(self):
        if len(self._data.shape) == 1:
            return self
        else:
            return ak._v2.contents.RegularArray(
                NumpyArray(
                    self._data.reshape((-1,) + self._data.shape[2:]),
                    None,
                    None,
                    nplike=self._nplike,
                ).toRegularArray(),
                self._data.shape[1],
                self._data.shape[0],
                self._identifier,
                self._parameters,
            )

    def maybe_to_nplike(self, nplike):
        return nplike.asarray(self._data)

    def _getitem_nothing(self):
        tmp = self._data[0:0]
        return NumpyArray(
            tmp.reshape((0,) + tmp.shape[2:]),
            self._range_identifier(0, 0),
            None,
            nplike=self._nplike,
        )

    def _getitem_at(self, where):
        try:
            out = self._data[where]
        except IndexError as err:
            raise NestedIndexError(self, where, str(err))

        if hasattr(out, "shape") and len(out.shape) != 0:
            return NumpyArray(out, None, None, nplike=self._nplike)
        else:
            return out

    def _getitem_range(self, where):
        start, stop, step = where.indices(len(self))
        assert step == 1

        try:
            out = self._data[where]
        except IndexError as err:
            raise NestedIndexError(self, where, str(err))

        return NumpyArray(
            out,
            self._range_identifier(start, stop),
            self._parameters,
            nplike=self._nplike,
        )

    def _getitem_field(self, where, only_fields=()):
        raise NestedIndexError(self, where, "not an array of records")

    def _getitem_fields(self, where, only_fields=()):
        if len(where) == 0:
            return self._getitem_range(slice(0, 0))
        raise NestedIndexError(self, where, "not an array of records")

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)
        try:
            nextdata = self._data[carry.data]
        except IndexError as err:
            if issubclass(exception, NestedIndexError):
                raise exception(self, carry.data, str(err))
            else:
                raise exception(str(err))

        return NumpyArray(
            nextdata,
            self._carry_identifier(carry, exception),
            self._parameters,
            nplike=self._nplike,
        )

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        if self._data.ndim == 1:
            raise NestedIndexError(
                self,
                ak._v2.contents.ListArray(slicestarts, slicestops, slicecontent),
                "too many jagged slice dimensions for array",
            )
        else:
            next = self.toRegularArray()
            return next._getitem_next_jagged(
                slicestarts, slicestops, slicecontent, tail
            )

    def _getitem_next(self, head, tail, advanced):
        nplike = self._nplike

        if head == ():
            return self

        elif isinstance(head, int):
            where = (slice(None), head) + tail

            try:
                out = self._data[where]
            except IndexError as err:
                raise NestedIndexError(self, (head,) + tail, str(err))

            if hasattr(out, "shape") and len(out.shape) != 0:
                return NumpyArray(out, None, None, nplike=nplike)
            else:
                return out

        elif isinstance(head, slice) or head is np.newaxis or head is Ellipsis:
            where = (slice(None), head) + tail
            try:
                out = self._data[where]
            except IndexError as err:
                raise NestedIndexError(self, (head,) + tail, str(err))
            out2 = NumpyArray(out, None, self._parameters, nplike=nplike)
            return out2

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif isinstance(head, ak._v2.index.Index64):
            if advanced is None:
                where = (slice(None), head.data) + tail
            else:
                where = (nplike.asarray(advanced.data), head.data) + tail

            try:
                out = self._data[where]
            except IndexError as err:
                raise NestedIndexError(self, (head,) + tail, str(err))

            return NumpyArray(out, None, self._parameters, nplike=nplike)

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            where = (slice(None), head) + tail
            try:
                out = self._data[where]
            except IndexError as err:
                raise NestedIndexError(self, (head,) + tail, str(err))
            out2 = NumpyArray(out, None, self._parameters, nplike=nplike)
            return out2

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            next = self.toRegularArray()
            return next._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def _localindex(self, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        elif len(self.shape) <= 1:
            raise np.AxisError(self, "'axis' out of range for localindex")
        else:
            return self.toRegularArray()._localindex(posaxis, depth)

    def contiguous(self):
        if self._data.ndim >= 1:
            return ak._v2.contents.NumpyArray(self.nplike.ascontiguousarray(self))
        else:
            return self

    def iscontiguous(self):
        x = self._data.dtype.itemsize

        for i in range(len(self.shape), 0, -1):  # FIXME: more Pythonic way to do it?
            if x != self.strides[i - 1]:
                return False
            else:
                x = x * self.shape[i - 1]

        return True

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if self.shape[0] == 0:
            return ak._v2.contents.NumpyArray(self.nplike.empty(0, np.int64))

        if len(self.shape) == 0:
            raise TypeError(
                "{0} attempting to argsort a scalar ".format(type(self).__name__)
            )
        elif len(self.shape) != 1 or not self.iscontiguous():
            contiguous_self = self if self.iscontiguous() else self.contiguous()
            return contiguous_self.toRegularArray()._argsort_next(
                negaxis,
                starts,
                shifts,
                parents,
                outlength,
                ascending,
                stable,
                kind,
                order,
            )

        else:
            nplike = self.nplike

            parents_length = len(parents)
            offsets_length = ak._v2.index.Index64.empty(1, nplike)
            self._handle_error(
                nplike[
                    "awkward_sorting_ranges_length",
                    offsets_length.dtype.type,
                    parents.dtype.type,
                ](
                    offsets_length.to(nplike),
                    parents.to(nplike),
                    parents_length,
                )
            )
            offsets_length = offsets_length[0]

            offsets = ak._v2.index.Index64.empty(offsets_length, nplike)
            self._handle_error(
                nplike[
                    "awkward_sorting_ranges",
                    offsets.dtype.type,
                    parents.dtype.type,
                ](
                    offsets.to(nplike),
                    offsets_length,
                    parents.to(nplike),
                    parents_length,
                )
            )

            nextcarry = ak._v2.index.Index64.empty(self.__len__(), nplike)
            self._handle_error(
                nplike[
                    "awkward_argsort",
                    nextcarry.dtype.type,
                    self._data.dtype.type,
                    offsets.dtype.type,
                ](
                    nextcarry.to(nplike),
                    self._data,
                    self.__len__(),
                    offsets.to(nplike),
                    offsets_length,
                    ascending,
                    stable,
                )
            )

            if shifts is not None:
                self._handle_error(
                    nplike[
                        "awkward_NumpyArray_rearrange_shifted",
                        nextcarry.dtype.type,
                        shifts.dtype.type,
                        offsets.dtype.type,
                        parents.dtype.type,
                        starts.dtype.type,
                    ](
                        nextcarry.to(nplike),
                        shifts.to(nplike),
                        len(shifts),
                        offsets.to(nplike),
                        offsets_length,
                        parents.to(nplike),
                        parents_length,
                        starts.to(nplike),
                        len(starts),
                    )
                )
            out = NumpyArray(nextcarry)
            return out

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        if len(self.shape) == 0:
            raise TypeError(
                "{0} attempting to sort a scalar ".format(type(self).__name__)
            )

        elif len(self.shape) != 1 or not self.iscontiguous():
            contiguous_self = self if self.iscontiguous() else self.contiguous()
            return contiguous_self.toRegularArray()._sort_next(
                negaxis,
                starts,
                parents,
                outlength,
                ascending,
                stable,
                kind,
                order,
            )

        else:
            nplike = self.nplike

            parents_length = len(parents)
            offsets_length = ak._v2.index.Index64.empty(1, nplike)
            self._handle_error(
                nplike[
                    "awkward_sorting_ranges_length",
                    offsets_length.dtype.type,
                    parents.dtype.type,
                ](
                    offsets_length.to(nplike),
                    parents.to(nplike),
                    parents_length,
                )
            )

            offsets = ak._v2.index.Index64.zeros(offsets_length[0], nplike)

            self._handle_error(
                nplike[
                    "awkward_sorting_ranges",
                    offsets.dtype.type,
                    parents.dtype.type,
                ](
                    offsets.to(nplike),
                    offsets_length[0],
                    parents.to(nplike),
                    parents_length,
                )
            )

            out = ak._v2.contents.NumpyArray(nplike.empty(len(self._data), self.dtype))
            self._handle_error(
                nplike[
                    "awkward_sort",
                    out._data.dtype.type,
                    self._data.dtype.type,
                    offsets.dtype.type,
                ](
                    out._data,
                    self._data,
                    self.shape[0],
                    offsets.to(nplike),
                    offsets_length[0],
                    parents_length,
                    ascending,
                    stable,
                )
            )
            return out

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        elif len(self.shape) <= 1:
            raise np.AxisError("'axis' out of range for combinations")
        else:
            return self.toRegularArray()._combinations(
                n, replacement, recordlookup, parameters, posaxis, depth
            )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        nplike = self.nplike

        out = reducer.apply(self, parents, outlength)

        if reducer.needs_position:
            if shifts is None:
                self._handle_error(
                    nplike[
                        "awkward_NumpyArray_reduce_adjust_starts_64",
                        out.data.dtype.type,
                        parents.dtype.type,
                        starts.dtype.type,
                    ](
                        out.data,
                        outlength,
                        parents.to(nplike),
                        starts.to(nplike),
                    )
                )
            else:
                self._handle_error(
                    nplike[
                        "awkward_NumpyArray_reduce_adjust_starts_shifts_64",
                        out.data.dtype.type,
                        parents.dtype.type,
                        starts.dtype.type,
                        shifts.dtype.type,
                    ](
                        out.data,
                        outlength,
                        parents.to(nplike),
                        starts.to(nplike),
                        shifts.to(nplike),
                    )
                )

        if mask:
            outmask = ak._v2.index.Index8.zeros(outlength, nplike)
            self._handle_error(
                nplike[
                    "awkward_NumpyArray_reduce_mask_ByteMaskedArray_64",
                    outmask.dtype.type,
                    parents.dtype.type,
                ](
                    outmask.to(nplike),
                    parents.to(nplike),
                    len(parents),
                    outlength,
                )
            )

            out = ak._v2.contents.ByteMaskedArray(
                outmask,
                out,
                False,
                None,
                None,
            )

        if keepdims:
            out = ak._v2.contents.RegularArray(
                out,
                1,
                len(self),
                None,
                None,
            )

        return out

    def _validityerror(self, path):
        if len(self.shape) == 0:
            return 'at {0} ("{1}"): shape is zero-dimensional'.format(path, type(self))
        for i in range(len(self.shape)):
            if self.shape[i] < 0:
                return 'at {0} ("{1}"): shape[{2}] < 0'.format(path, type(self), i)
        for i in range(len(self.strides)):
            if self.strides[i] % self.dtype.itemsize != 0:
                return 'at {0} ("{1}"): shape[{2}] % itemsize != 0'.format(
                    path, type(self), i
                )
        return ""



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


@pytest.mark.skipif(
    ak._util.py27,
    reason="np.ma.zeros(..., structured dtype) fails for NumPy " + str(np.__version__),
)
def test():
    a = ak.to_numpy(ak.Array({"A": [1, 2, 3], "B": [4, None, 5]}))
    assert a["A"].data.tolist() == [1, 2, 3]
    assert a["A"].mask.tolist() == [False, False, False]
    assert a["B"].data[0] == 4
    assert a["B"].data[2] == 5
    assert a["B"].mask.tolist() == [False, True, False]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_numpy_array():
    x = np.arange(12)
    y = ak.is_none(x)

    assert y.tolist() == [False] * 12


def test_awkward_from_numpy_array():
    x = np.arange(12)
    y = ak.from_numpy(x)
    z = ak.is_none(y)

    assert z.tolist() == [False] * 12



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_getitem():
    a = ak.from_json("[]")
    a = ak.from_json("[[], [[], []], [[], [], []]]")

    a = v1_to_v2(a.layout)

    assert ak.to_list(a[2]) == [[], [], []]
    assert a.typetracer[2].form == a[2].form

    assert ak.to_list(a[2, 1]) == []
    assert a.typetracer[2, 1].form == a[2, 1].form
    with pytest.raises(ValueError):
        a[2, 1, 0]
    assert ak.to_list(a[2, 1][()]) == []
    assert a.typetracer[2, 1][()].form == a[2, 1][()].form
    with pytest.raises(IndexError):
        a[2, 1][0]
    assert ak.to_list(a[2, 1][100:200]) == []
    assert a.typetracer[2, 1][100:200].form == a[2, 1][100:200].form
    assert ak.to_list(a[2, 1, 100:200]) == []
    assert a.typetracer[2, 1, 100:200].form == a[2, 1, 100:200].form
    assert ak.to_list(a[2, 1][np.array([], dtype=int)]) == []
    assert (
        a.typetracer[2, 1][np.array([], dtype=int)].form
        == a[2, 1][np.array([], dtype=int)].form
    )
    assert ak.to_list(a[2, 1, np.array([], dtype=int)]) == []
    assert (
        a.typetracer[2, 1, np.array([], dtype=int)].form
        == a[2, 1, np.array([], dtype=int)].form
    )
    with pytest.raises(ValueError):
        a[2, 1, np.array([0], dtype=int)]
    with pytest.raises(IndexError):
        a[2, 1][100:200, 0]
    with pytest.raises(IndexError):
        a[2, 1][100:200, 200:300]

    assert ak.to_list(a[2, 1][100:200, np.array([], dtype=int)]) == []
    assert (
        a.typetracer[2, 1][100:200, np.array([], dtype=int)].form
        == a[2, 1][100:200, np.array([], dtype=int)].form
    )

    assert ak.to_list(a[1:, 1:]) == [[[]], [[], []]]
    assert a.typetracer[1:, 1:].form == a[1:, 1:].form
    with pytest.raises(ValueError):
        a[1:, 1:, 0]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import json

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    f = ak.forms.Form.fromjson(
        json.dumps(
            {
                "class": "NumpyArray",
                "itemsize": 8,
                "format": "d",
                "primitive": "float64",
                "parameters": {"thing": 1.23},
            }
        )
    )
    assert json.loads(repr(f)) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
        "parameters": {"thing": 1.23},  # not 1 (int)
    }



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


class CategoricalBehavior(ak.highlevel.Array):
    __name__ = "Array"


ak.behavior["categorical"] = CategoricalBehavior


class _HashableDict(object):
    def __init__(self, obj):
        self.keys = tuple(sorted(obj))
        self.values = tuple(_hashable(obj[k]) for k in self.keys)
        self.hash = hash((_HashableDict,) + self.keys, self.values)

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (
            isinstance(other, _HashableDict)
            and self.keys == other.keys
            and self.values == other.values
        )


class _HashableList(object):
    def __init__(self, obj):
        self.values = tuple(obj)
        self.hash = hash((_HashableList,) + self.values)

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return isinstance(other, _HashableList) and self.values == other.values


def _hashable(obj):
    if isinstance(obj, dict):
        return _HashableDict(obj)
    elif isinstance(obj, tuple):
        return tuple(_hashable(x) for x in obj)
    elif isinstance(obj, list):
        return _HashableList(obj)
    else:
        return obj


def _categorical_equal(one, two):
    behavior = ak._util.behaviorof(one, two)

    one, two = one.layout, two.layout

    assert isinstance(one, ak._util.indexedtypes + ak._util.indexedoptiontypes)
    assert isinstance(two, ak._util.indexedtypes + ak._util.indexedoptiontypes)
    assert one.parameter("__array__") == "categorical"
    assert two.parameter("__array__") == "categorical"

    one_index = ak.nplike.numpy.asarray(one.index)
    two_index = ak.nplike.numpy.asarray(two.index)
    one_content = ak._util.wrap(one.content, behavior)
    two_content = ak._util.wrap(two.content, behavior)

    if len(one_content) == len(two_content) and ak.operations.reducers.all(
        one_content == two_content, axis=None
    ):
        one_mapped = one_index

    else:
        one_list = ak.operations.convert.to_list(one_content)
        two_list = ak.operations.convert.to_list(two_content)
        one_hashable = [_hashable(x) for x in one_list]
        two_hashable = [_hashable(x) for x in two_list]
        two_lookup = {x: i for i, x in enumerate(two_hashable)}

        one_to_two = ak.nplike.numpy.empty(len(one_hashable) + 1, dtype=np.int64)
        for i, x in enumerate(one_hashable):
            one_to_two[i] = two_lookup.get(x, len(two_hashable))
        one_to_two[-1] = -1

        one_mapped = one_to_two[one_index]

    out = one_mapped == two_index
    return ak._util.wrap(ak.layout.NumpyArray(out), ak._util.behaviorof(one, two))


ak.behavior[ak.nplike.numpy.equal, "categorical", "categorical"] = _categorical_equal


def _apply_ufunc(ufunc, method, inputs, kwargs):
    nextinputs = []
    for x in inputs:
        if isinstance(x, ak.highlevel.Array) and isinstance(
            x.layout, ak._util.indexedtypes
        ):
            nextinputs.append(
                ak.highlevel.Array(x.layout.project(), behavior=ak._util.behaviorof(x))
            )
        else:
            nextinputs.append(x)

    return getattr(ufunc, method)(*nextinputs, **kwargs)


ak.behavior[ak.nplike.numpy.ufunc, "categorical"] = _apply_ufunc


def is_categorical(array):
    """
    Args:
        array: A possibly-categorical Awkward Array.

    If the `array` is categorical (contains #ak.layout.IndexedArray or
    #ak.layout.IndexedOptionArray labeled with parameter
    `"__array__" = "categorical"`), then this function returns True;
    otherwise, it returns False.

    See also #ak.categories, #ak.to_categorical, #ak.from_categorical.
    """

    layout = ak.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    return layout.purelist_parameter("__array__") == "categorical"


def categories(array, highlevel=True):
    """
    Args:
        array: A possibly-categorical Awkward Array.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.

    If the `array` is categorical (contains #ak.layout.IndexedArray or
    #ak.layout.IndexedOptionArray labeled with parameter
    `"__array__" = "categorical"`), then this function returns its categories.

    See also #ak.is_categorical, #ak.to_categorical, #ak.from_categorical.
    """

    output = [None]

    def getfunction(layout):
        if layout.parameter("__array__") == "categorical":
            output[0] = layout.content
            return lambda: layout

        else:
            return None

    layout = ak.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    ak._util.recursively_apply(layout, getfunction, pass_depth=False)

    if output[0] is None:
        return None
    elif highlevel:
        return ak._util.wrap(output[0], ak._util.behaviorof(array))
    else:
        return output[0]


def to_categorical(array, highlevel=True):
    """
    Args:
        array: Data convertible to an Awkward Array
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.

    Creates a categorical dataset, which has the following properties:

       * only distinct values (categories) are stored in their entirety,
       * pointers to those distinct values are represented by integers
         (an #ak.layout.IndexedArray or #ak.layout.IndexedOptionArray
         labeled with parameter `"__array__" = "categorical"`.

    This is equivalent to R's "factor", Pandas's "categorical", and
    Arrow/Parquet's "dictionary encoding." It differs from generic uses of
    #ak.layout.IndexedArray and #ak.layout.IndexedOptionArray in Awkward
    Arrays by the guarantee of no duplicate categories and the `"categorical"`
    parameter.

        >>> array = ak.Array([["one", "two", "three"], [], ["three", "two"]])
        >>> categorical = ak.to_categorical(array)
        >>> categorical
        <Array [['one', 'two', ... 'three', 'two']] type='3 * var * categorical[type=str...'>
        >>> ak.type(categorical)
        3 * var * categorical[type=string]
        >>> ak.to_list(categorical) == ak.to_list(array)
        True
        >>> ak.categories(categorical)
        <Array ['one', 'two', 'three'] type='3 * string'>
        >>> ak.is_categorical(categorical)
        True
        >>> ak.from_categorical(categorical)
        <Array [['one', 'two', ... 'three', 'two']] type='3 * var * string'>

    This function descends through nested lists, but not into the fields of
    records, so records can be categories. To make categorical record
    fields, split up the record, apply this function to each desired field,
    and #ak.zip the results together.

        >>> records = ak.Array([
        ...     {"x": 1.1, "y": "one"},
        ...     {"x": 2.2, "y": "two"},
        ...     {"x": 3.3, "y": "three"},
        ...     {"x": 2.2, "y": "two"},
        ...     {"x": 1.1, "y": "one"}
        ... ])
        >>> records
        <Array [{x: 1.1, y: 'one'}, ... y: 'one'}] type='5 * {"x": float64, "y": string}'>
        >>> categorical_records = ak.zip({
        ...     "x": ak.to_categorical(records["x"]),
        ...     "y": ak.to_categorical(records["y"]),
        ... })
        >>> categorical_records
        <Array [{x: 1.1, y: 'one'}, ... y: 'one'}] type='5 * {"x": categorical[type=floa...'>
        >>> ak.type(categorical_records)
        5 * {"x": categorical[type=float64], "y": categorical[type=string]}
        >>> ak.to_list(categorical_records) == ak.to_list(records)
        True

    The check for uniqueness is currently implemented in a Python loop, so
    conversion to categorical should be regarded as expensive. (This can
    change, but it would always be an _n log(n)_ operation.)

    See also #ak.is_categorical, #ak.categories, #ak.from_categorical.
    """

    def getfunction(layout):
        if layout.purelist_depth == 1:
            if isinstance(layout, ak._util.optiontypes):
                layout = layout.simplify()

            if isinstance(layout, ak._util.indexedoptiontypes):
                content = layout.content
                cls = ak.layout.IndexedOptionArray64
            elif isinstance(layout, ak._util.indexedtypes):
                content = layout.content
                cls = ak.layout.IndexedArray64
            elif isinstance(layout, ak._util.optiontypes):
                content = layout.content
                cls = ak.layout.IndexedOptionArray64
            else:
                content = layout
                cls = ak.layout.IndexedArray64

            content_list = ak.operations.convert.to_list(content)
            hashable = [_hashable(x) for x in content_list]

            lookup = {}
            is_first = ak.nplike.numpy.empty(len(hashable), dtype=np.bool_)
            mapping = ak.nplike.numpy.empty(len(hashable), dtype=np.int64)
            for i, x in enumerate(hashable):
                if x in lookup:
                    is_first[i] = False
                    mapping[i] = lookup[x]
                else:
                    is_first[i] = True
                    lookup[x] = j = len(lookup)
                    mapping[i] = j

            if isinstance(layout, ak._util.indexedoptiontypes):
                original_index = ak.nplike.numpy.asarray(layout.index)
                index = mapping[original_index]
                index[original_index < 0] = -1
                index = ak.layout.Index64(index)

            elif isinstance(layout, ak._util.indexedtypes):
                original_index = ak.nplike.numpy.asarray(layout.index)
                index = ak.layout.Index64(mapping[original_index])

            elif isinstance(layout, ak._util.optiontypes):
                mask = ak.nplike.numpy.asarray(layout.bytemask())
                mapping[mask.view(np.bool_)] = -1
                index = ak.layout.Index64(mapping)

            else:
                index = ak.layout.Index64(mapping)

            out = cls(index, content[is_first], parameters={"__array__": "categorical"})
            return lambda: out

        else:
            return None

    layout = ak.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    out = ak._util.recursively_apply(layout, getfunction, pass_depth=False)
    if highlevel:
        return ak._util.wrap(out, ak._util.behaviorof(array))
    else:
        return out


def from_categorical(array, highlevel=True):
    """
    Args:
        array: Awkward Array from which to remove the 'categorical' parameter.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.

    This function replaces categorical data with non-categorical data (by
    removing the label that declares it as such).

    This is a metadata-only operation; the running time does not scale with the
    size of the dataset. (Conversion to categorical is expensive; conversion
    from categorical is cheap.)

    See also #ak.is_categorical, #ak.categories, #ak.to_categorical,
    #ak.from_categorical.
    """

    def getfunction(layout):
        if layout.parameter("__array__") == "categorical":
            out = ak.operations.structure.with_parameter(
                layout, "__array__", None, highlevel=False
            )
            return lambda: out

        else:
            return None

    layout = ak.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    out = ak._util.recursively_apply(layout, getfunction, pass_depth=False)
    if highlevel:
        return ak._util.wrap(out, ak._util.behaviorof(array))
    else:
        return out


__all__ = [
    x for x in list(globals()) if not x.startswith("_") and x not in ("ak", "np")
]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


class Reducer(object):
    needs_position = False

    @classmethod
    def return_dtype(cls, given_dtype):
        if (
            given_dtype == np.bool_
            or given_dtype == np.int8
            or given_dtype == np.int16
            or given_dtype == np.int32
        ):
            if ak._util.win or ak._util.bits32:
                return np.int32
            else:
                return np.int64

        if (
            given_dtype == np.uint8
            or given_dtype == np.uint16
            or given_dtype == np.uint32
        ):
            if ak._util.win or ak._util.bits32:
                return np.uint32
            else:
                return np.uint64

        return given_dtype


class ArgMin(Reducer):
    name = "argmin"

    needs_position = True
    preferred_dtype = np.int64

    @classmethod
    def apply(cls, array, parents, outlength):
        result = ak._v2.index.Index64.empty(
            outlength, array.nplike, dtype=cls.preferred_dtype
        )
        array._handle_error(
            array.nplike[
                "awkward_reduce_argmin",
                result.dtype.type,
                array.data.dtype.type,
                parents.dtype.type,
            ](
                result.to(array.nplike),
                array.data,
                parents.to(array.nplike),
                len(parents),
                outlength,
            )
        )
        return ak._v2.contents.NumpyArray(result)


class ArgMax(Reducer):
    name = "argmax"

    needs_position = True
    preferred_dtype = np.int64

    @classmethod
    def apply(cls, array, parents, outlength):
        result = ak._v2.index.Index64.empty(
            outlength, array.nplike, dtype=cls.preferred_dtype
        )
        array._handle_error(
            array.nplike[
                "awkward_reduce_argmax",
                result.dtype.type,
                array._data.dtype.type,
                parents.dtype.type,
            ](
                result.to(array.nplike),
                array._data,
                parents.to(array.nplike),
                len(parents),
                outlength,
            )
        )
        return ak._v2.contents.NumpyArray(result)


class Count(Reducer):
    name = "count"
    preferred_dtype = np.float64

    @classmethod
    def apply(cls, array, parents, outlength):
        result = ak._v2.index.Index64.empty(outlength, array.nplike)
        array._handle_error(
            array.nplike[
                "awkward_reduce_count_64", result.dtype.type, parents.dtype.type
            ](
                result.to(array.nplike),
                parents.to(array.nplike),
                len(parents),
                outlength,
            )
        )
        return ak._v2.contents.NumpyArray(result)


class CountNonzero(Reducer):
    name = "count_nonzero"
    preferred_dtype = np.float64

    @classmethod
    def apply(cls, array, parents, outlength):
        result = ak._v2.index.Index64.empty(outlength, array.nplike)
        array._handle_error(
            array.nplike[
                "awkward_reduce_countnonzero",
                result.dtype.type,
                array._data.dtype.type,
                parents.dtype.type,
            ](
                result.to(array.nplike),
                array._data,
                parents.to(array.nplike),
                len(parents),
                outlength,
            )
        )
        return ak._v2.contents.NumpyArray(result)


class Sum(Reducer):
    name = "sum"
    preferred_dtype = np.float64

    @classmethod
    def apply(cls, array, parents, outlength):
        result = ak._v2.contents.NumpyArray(
            array.nplike.empty(outlength, dtype=cls.return_dtype(array.dtype))
        )
        if array.dtype == np.bool_:
            if result.dtype == np.int64 or result.dtype == np.uint64:
                array._handle_error(
                    array.nplike[
                        "awkward_reduce_sum_int64_bool_64",
                        np.int64,
                        array.dtype.type,
                        parents.dtype.type,
                    ](
                        result._data,
                        array._data,
                        parents.to(array.nplike),
                        len(parents),
                        outlength,
                    )
                )
            elif result.dtype == np.int32 or result.dtype == np.uint32:
                array._handle_error(
                    array.nplike[
                        "awkward_reduce_sum_int32_bool_64",
                        np.int32,
                        array.dtype.type,
                        parents.dtype.type,
                    ](
                        result._data,
                        array._data,
                        parents.to(array.nplike),
                        len(parents),
                        outlength,
                    )
                )
            else:
                raise NotImplementedError
        else:
            array._handle_error(
                array.nplike[
                    "awkward_reduce_sum",
                    result.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result._data,
                    array._data,
                    parents.to(array.nplike),
                    len(parents),
                    outlength,
                )
            )
        return ak._v2.contents.NumpyArray(result)


class Prod(Reducer):
    name = "prod"
    preferred_dtype = np.int64

    @classmethod
    def apply(cls, array, parents, outlength):
        result = ak._v2.contents.NumpyArray(
            array.nplike.empty(outlength, dtype=cls.return_dtype(array.dtype))
        )
        if array.dtype == np.bool_:
            array._handle_error(
                array.nplike[
                    "awkward_reduce_prod_bool",
                    array.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result._data,
                    array._data,
                    parents.to(array.nplike),
                    len(parents),
                    outlength,
                )
            )
        else:
            array._handle_error(
                array.nplike[
                    "awkward_reduce_prod",
                    result.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result._data,
                    array._data,
                    parents.to(array.nplike),
                    len(parents),
                    outlength,
                )
            )
        return ak._v2.contents.NumpyArray(result)


class Any(Reducer):
    name = "any"
    preferred_dtype = np.bool_

    @classmethod
    def apply(cls, array, parents, outlength):
        result = ak._v2.contents.NumpyArray(
            array.nplike.empty(outlength, dtype=cls.preferred_dtype)
        )
        array._handle_error(
            array.nplike[
                "awkward_reduce_sum_bool",
                result.dtype.type,
                array._data.dtype.type,
                parents.dtype.type,
            ](
                result._data,
                array._data,
                parents.to(array.nplike),
                len(parents),
                outlength,
            )
        )
        return ak._v2.contents.NumpyArray(result)


class All(Reducer):
    name = "all"
    preferred_dtype = np.bool_

    @classmethod
    def apply(cls, array, parents, outlength):
        result = ak._v2.contents.NumpyArray(
            array.nplike.empty(outlength, dtype=cls.preferred_dtype)
        )
        array._handle_error(
            array.nplike[
                "awkward_reduce_prod_bool",
                result.dtype.type,
                array._data.dtype.type,
                parents.dtype.type,
            ](
                result._data,
                array._data,
                parents.to(array.nplike),
                len(parents),
                outlength,
            )
        )
        return ak._v2.contents.NumpyArray(result)


class Min(Reducer):
    name = "min"
    preferred_dtype = np.float64
    initial = None

    def __init__(self, initial):
        type(self).initial = initial

    def __del__(self):
        type(self).initial = None

    @staticmethod
    def _min_initial(initial, type):
        if initial is None:
            if (
                type == np.int8
                or type == np.int16
                or type == np.int32
                or type == np.int64
                or type == np.uint8
                or type == np.uint16
                or type == np.uint32
                or type == np.uint64
            ):
                return np.iinfo(type).max
            else:
                return np.inf

        return initial

    @classmethod
    def apply(cls, array, parents, outlength):
        dtype = array.dtype
        result = ak._v2.contents.NumpyArray(array.nplike.empty(outlength, dtype))

        if array.dtype == np.bool_:
            array._handle_error(
                array.nplike[
                    "awkward_reduce_prod_bool",
                    result.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result._data,
                    array._data,
                    parents.to(array.nplike),
                    len(parents),
                    outlength,
                )
            )
        else:
            array._handle_error(
                array.nplike[
                    "awkward_reduce_min",
                    result.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result._data,
                    array._data,
                    parents.to(array.nplike),
                    len(parents),
                    outlength,
                    cls._min_initial(cls.initial, dtype.type),
                )
            )

        return ak._v2.contents.NumpyArray(result)


class Max(Reducer):
    name = "max"
    preferred_dtype = np.float64
    initial = None

    def __init__(self, initial):
        type(self).initial = initial

    def __del__(self):
        type(self).initial = None

    @staticmethod
    def _max_initial(initial, type):
        if initial is None:
            if (
                type == np.int8
                or type == np.int16
                or type == np.int32
                or type == np.int64
                or type == np.uint8
                or type == np.uint16
                or type == np.uint32
                or type == np.uint64
            ):
                return np.iinfo(type).min
            else:
                return -np.inf

        return initial

    @classmethod
    def apply(cls, array, parents, outlength):
        dtype = array.dtype
        result = ak._v2.contents.NumpyArray(array.nplike.empty(outlength, dtype))

        if array.dtype == np.bool_:
            array._handle_error(
                array.nplike[
                    "awkward_reduce_sum_bool",
                    result.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result._data,
                    array._data,
                    parents.to(array.nplike),
                    len(parents),
                    outlength,
                )
            )
        else:
            array._handle_error(
                array.nplike[
                    "awkward_reduce_max",
                    result.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result._data,
                    array._data,
                    parents.to(array.nplike),
                    len(parents),
                    outlength,
                    cls._max_initial(cls.initial, dtype.type),
                )
            )
        return ak._v2.contents.NumpyArray(result)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    nparray = ak.layout.NumpyArray(np.array([1, 2, 3], dtype=np.int64))
    listarray = ak.layout.ListArray64(
        ak.layout.Index64(np.array([0], dtype=np.int64)),
        ak.layout.Index64(np.array([3], dtype=np.int64)),
        nparray,
    )
    indexedarray = ak.layout.IndexedArray64(ak.layout.Index64([]), listarray)

    cart = ak.cartesian([indexedarray, indexedarray], nested=True)
    assert str(cart.type) == "0 * var * var * (int64, int64)"
    assert ak.to_list(cart) == []

    cart = ak.cartesian([indexedarray, indexedarray], nested=False)
    assert str(cart.type) == "0 * var * (int64, int64)"
    assert ak.to_list(cart) == []



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import copy

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    np_array = np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    one = ak.Array(np_array)

    np_array[1] = 999
    assert ak.to_list(one) == [0.0, 999, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]

    two = copy.copy(one)
    np_array[3] = 123
    assert ak.to_list(two) == [0.0, 999, 2.2, 123, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]

    three = copy.deepcopy(two)
    four = np.copy(two)
    np_array[5] = 321
    assert ak.to_list(three) == [0.0, 999, 2.2, 123, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]
    assert ak.to_list(four) == [0.0, 999, 2.2, 123, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]

    assert ak.to_list(copy.deepcopy(ak.Array([[1, 2, 3], [], [4, 5]]))) == [
        [1, 2, 3],
        [],
        [4, 5],
    ]

    assert ak.to_list(copy.deepcopy(ak.Record({"one": 1, "two": 2.2}))) == ak.to_list(
        copy.deepcopy(ak.Record({"one": 1, "two": 2.2}))
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array(
        [[{"a": 1, "b": [1, 2, 3]}], [{"a": 1, "b": [4, 5]}, {"a": 4, "b": [2]}]]
    )
    array_new = ak.Array(
        ak.layout.ListOffsetArray64(
            array.layout.offsets,
            ak.layout.RecordArray(
                {
                    "a": array.layout.content["a"],
                    "b": ak.layout.ListArray64(
                        array.layout.content["b"].offsets[:-1],
                        array.layout.content["b"].offsets[1:],
                        array.layout.content["b"].content,
                    ),
                }
            ),
        )
    )
    form, length, container = ak.to_buffers(array_new)
    reconstituted = ak.from_buffers(form, length, container, lazy=True)
    assert reconstituted.tolist() == array_new.tolist()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_bool_sort():
    array = ak.layout.NumpyArray(np.array([True, False, True, False, False]))
    array = v1_to_v2(array)
    assert ak.to_list(array.sort(axis=0, ascending=True, stable=False)) == [
        False,
        False,
        False,
        True,
        True,
    ]


def test_keep_None_in_place_test():
    array = ak.Array([[3, 2, 1], [], None, [4, 5]])
    array = v1_to_v2(array.layout)

    assert ak.to_list(array.argsort(axis=1)) == [
        [2, 1, 0],
        [],
        None,
        [0, 1],
    ]

    assert ak.to_list(array.sort(axis=1)) == [
        [1, 2, 3],
        [],
        None,
        [4, 5],
    ]


@pytest.mark.skip(
    reason="FIXME: slicing might infinite-loop before simplify_optiontype is implemented"
)
def test_slicing_FIXME():
    # awkward/_v2/_slicing.py:218:
    array = ak.Array([[3, 2, 1], [], None, [4, 5]])
    array = v1_to_v2(array.layout)

    assert ak.to_list(array[array.argsort(axis=1)]) == ak.to_list(array.sort(axis=1))


def test_EmptyArray():
    array = ak.layout.EmptyArray()
    array = v1_to_v2(array)

    assert ak.to_list(array.sort()) == []
    assert ak.to_list(array.argsort()) == []

    array2 = ak.Array([[], [], []])
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(array2.argsort()) == [[], [], []]


@pytest.mark.skip(reason="FIXME: 'NumpyForm' object has no attribute 'type'")
def test_EmptyArray_type_FIXME():
    array = ak.layout.EmptyArray()
    array = v1_to_v2(array)

    assert str(array.sort().form.type) == "0 * float64"
    assert str(array.argsort().form.type) == "0 * int64"

    array2 = ak.Array([[], [], []])
    array2 = v1_to_v2(array2.layout)
    assert str(array2.argsort().form.type) == "3 * var * int64"


def test_NumpyArray():
    array = ak.layout.NumpyArray(np.array([3.3, 2.2, 1.1, 5.5, 4.4]))
    array = v1_to_v2(array)

    assert ak.to_list(array.argsort(axis=0, ascending=True, stable=False)) == [
        2,
        1,
        0,
        4,
        3,
    ]
    assert ak.to_list(array.argsort(axis=0, ascending=False, stable=False)) == [
        3,
        4,
        0,
        1,
        2,
    ]

    assert ak.to_list(array.sort(axis=0, ascending=True, stable=False)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]
    assert ak.to_list(array.sort(axis=0, ascending=False, stable=False)) == [
        5.5,
        4.4,
        3.3,
        2.2,
        1.1,
    ]

    array2 = ak.layout.NumpyArray(np.array([[3.3, 2.2, 4.4], [1.1, 5.5, 3.3]]))
    array2 = v1_to_v2(array2)

    assert ak.to_list(array2.sort(axis=1, ascending=True, stable=False)) == ak.to_list(
        np.sort(array2, axis=1)
    )
    assert ak.to_list(array2.sort(axis=0, ascending=True, stable=False)) == ak.to_list(
        np.sort(array2, axis=0)
    )

    assert ak.to_list(
        array2.argsort(axis=1, ascending=True, stable=False)
    ) == ak.to_list(np.argsort(array2, 1))

    assert ak.to_list(
        array2.argsort(axis=0, ascending=True, stable=False)
    ) == ak.to_list(np.argsort(array2, 0))

    with pytest.raises(ValueError) as err:
        array2.sort(axis=2, ascending=True, stable=False)
    assert str(err.value).startswith(
        "axis=2 exceeds the depth of the nested list structure (which is 2)"
    )


def test_IndexedOptionArray():
    array = ak.Array(
        [
            [None, None, 2.2, 1.1, 3.3],
            [None, None, None],
            [4.4, None, 5.5],
            [5.5, None, None],
            [-4.4, -5.5, -6.6],
        ]
    )
    array = v1_to_v2(array.layout)

    assert ak.to_list(array.sort(axis=0, ascending=True, stable=False)) == [
        [-4.4, -5.5, -6.6, 1.1, 3.3],
        [4.4, None, 2.2],
        [5.5, None, 5.5],
        [None, None, None],
        [None, None, None],
    ]

    assert ak.to_list(array.sort(axis=1, ascending=True, stable=False)) == [
        [1.1, 2.2, 3.3, None, None],
        [None, None, None],
        [4.4, 5.5, None],
        [5.5, None, None],
        [-6.6, -5.5, -4.4],
    ]

    assert ak.to_list(array.sort(axis=1, ascending=False, stable=True)) == [
        [3.3, 2.2, 1.1, None, None],
        [None, None, None],
        [5.5, 4.4, None],
        [5.5, None, None],
        [-4.4, -5.5, -6.6],
    ]

    assert ak.to_list(array.sort(axis=1, ascending=False, stable=False)) == [
        [3.3, 2.2, 1.1, None, None],
        [None, None, None],
        [5.5, 4.4, None],
        [5.5, None, None],
        [-4.4, -5.5, -6.6],
    ]

    assert ak.to_list(array.argsort(axis=0, ascending=True, stable=True)) == [
        [4, 4, 4, 0, 0],
        [2, 0, 0],
        [3, 1, 2],
        [0, 2, 1],
        [1, 3, 3],
    ]

    assert ak.to_list(array.argsort(axis=0, ascending=True, stable=False)) == [
        [4, 4, 4, 0, 0],
        [2, 0, 0],
        [3, 1, 2],
        [0, 2, 1],
        [1, 3, 3],
    ]

    assert ak.to_list(array.argsort(axis=0, ascending=False, stable=True)) == [
        [3, 4, 2, 0, 0],
        [2, 0, 0],
        [4, 1, 4],
        [0, 2, 1],
        [1, 3, 3],
    ]
    assert ak.to_list(array.argsort(axis=0, ascending=False, stable=False)) == [
        [3, 4, 2, 0, 0],
        [2, 0, 0],
        [4, 1, 4],
        [0, 2, 1],
        [1, 3, 3],
    ]

    assert ak.to_list(array.argsort(axis=1, ascending=True, stable=True)) == [
        [3, 2, 4, 0, 1],
        [0, 1, 2],
        [0, 2, 1],
        [0, 1, 2],
        [2, 1, 0],
    ]

    assert ak.to_list(array.argsort(axis=1, ascending=True, stable=False)) == [
        [3, 2, 4, 0, 1],
        [0, 1, 2],
        [0, 2, 1],
        [0, 1, 2],
        [2, 1, 0],
    ]

    assert ak.to_list(array.argsort(axis=1, ascending=False, stable=True)) == [
        [4, 2, 3, 0, 1],
        [0, 1, 2],
        [2, 0, 1],
        [0, 1, 2],
        [0, 1, 2],
    ]

    array2 = ak.Array([None, None, 1, -1, 30])
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(array2.argsort(axis=0, ascending=True, stable=True)) == [
        3,
        2,
        4,
        0,
        1,
    ]

    array3 = ak.Array(
        [[2.2, 1.1, 3.3], [], [4.4, 5.5], [5.5], [-4.4, -5.5, -6.6]]
    ).layout
    array3 = v1_to_v2(array3)

    assert ak.to_list(array3.sort(axis=1, ascending=False, stable=False)) == [
        [3.3, 2.2, 1.1],
        [],
        [5.5, 4.4],
        [5.5],
        [-4.4, -5.5, -6.6],
    ]

    assert ak.to_list(array3.sort(axis=0, ascending=True, stable=False)) == [
        [-4.4, -5.5, -6.6],
        [],
        [2.2, 1.1],
        [4.4],
        [5.5, 5.5, 3.3],
    ]


def test_IndexedArray():
    content = ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    index1 = ak.layout.Index32(np.array([1, 2, 3, 4], dtype=np.int32))
    indexedarray1 = ak.layout.IndexedArray32(index1, content)
    indexedarray1 = v1_to_v2(indexedarray1)

    assert ak.to_list(indexedarray1.argsort(axis=0, ascending=True, stable=False)) == [
        0,
        1,
        2,
        3,
    ]

    index2 = ak._v2.index.Index(np.array([1, 2, 3], dtype=np.int64))
    indexedarray2 = ak._v2.contents.IndexedArray(index2, indexedarray1)

    assert ak.to_list(indexedarray2.sort(axis=0, ascending=False, stable=False)) == [
        5.5,
        4.4,
        3.3,
    ]

    index3 = ak._v2.index.Index32(np.array([1, 2], dtype=np.int32))
    indexedarray3 = ak._v2.contents.IndexedArray(index3, indexedarray2)

    assert ak.to_list(indexedarray3.sort(axis=0, ascending=True, stable=False)) == [
        4.4,
        5.5,
    ]


def test_3d():
    array = ak.layout.NumpyArray(
        np.array(
            [
                # axis 2:    0       1       2       3       4         # axis 1:
                [
                    [1.1, 2.2, 3.3, 4.4, 5.5],  # 0
                    [6.6, 7.7, 8.8, 9.9, 10.10],  # 1
                    [11.11, 12.12, 13.13, 14.14, 15.15],
                ],  # 2
                [
                    [-1.1, -2.2, -3.3, -4.4, -5.5],  # 3
                    [-6.6, -7.7, -8.8, -9.9, -10.1],  # 4
                    [-11.11, -12.12, -13.13, -14.14, -15.15],
                ],
            ]
        )
    )  # 5
    sorted = ak.argsort(array, axis=1, ascending=True, stable=False)
    assert ak.to_list(sorted) == ak.to_list(np.argsort(array, 1))

    array = v1_to_v2(array)

    sorted = array.argsort(axis=2, ascending=True, stable=False)
    assert ak.to_list(sorted) == ak.to_list(np.argsort(array, 2))

    sorted = array.sort(axis=2, ascending=True, stable=False)
    assert ak.to_list(sorted) == ak.to_list(np.sort(array, 2))

    sorted = array.argsort(axis=1, ascending=True, stable=False)

    assert ak.to_list(sorted) == ak.to_list(np.argsort(array, 1))

    sorted = array.sort(axis=1, ascending=True, stable=False)
    assert ak.to_list(sorted) == ak.to_list(np.sort(array, 1))

    sorted = array.sort(axis=1, ascending=False, stable=False)
    assert ak.to_list(sorted) == [
        [
            [11.11, 12.12, 13.13, 14.14, 15.15],
            [6.6, 7.7, 8.8, 9.9, 10.1],
            [1.1, 2.2, 3.3, 4.4, 5.5],
        ],
        [
            [-1.1, -2.2, -3.3, -4.4, -5.5],
            [-6.6, -7.7, -8.8, -9.9, -10.1],
            [-11.11, -12.12, -13.13, -14.14, -15.15],
        ],
    ]

    sorted = array.sort(axis=0, ascending=True, stable=False)
    assert ak.to_list(sorted) == ak.to_list(np.sort(array, 0))

    assert ak.to_list(
        array.argsort(axis=0, ascending=True, stable=False)
    ) == ak.to_list(np.argsort(array, 0))


def test_ByteMaskedArray():
    content = ak.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    mask = ak.layout.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.layout.ByteMaskedArray(mask, content, valid_when=False)
    array = v1_to_v2(array)
    sorted = array.argsort(axis=0, ascending=True, stable=False)
    assert ak.to_list(sorted) == [
        [0, 0, 0],
        [],
        [2, 2, 2, 2],
        None,
        None,
    ]

    sorted = array.sort(axis=0, ascending=True, stable=False)
    assert ak.to_list(sorted) == [
        [0.0, 1.1, 2.2],
        [],
        [6.6, 7.7, 8.8, 9.9],
        None,
        None,
    ]

    assert ak.to_list(array.sort(axis=0, ascending=False, stable=False)) == [
        [6.6, 7.7, 8.8],
        [],
        [0.0, 1.1, 2.2, 9.9],
        None,
        None,
    ]

    assert ak.to_list(array.argsort(axis=1, ascending=True, stable=False)) == [
        [0, 1, 2],
        [],
        None,
        None,
        [0, 1, 2, 3],
    ]

    assert ak.to_list(array.sort(1, False, False)) == [
        [2.2, 1.1, 0.0],
        [],
        None,
        None,
        [9.9, 8.8, 7.7, 6.6],
    ]


def test_UnionArray():
    content0 = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    content1 = ak.from_iter(
        [["one"], ["two"], ["three"], ["four"], ["five"]], highlevel=False
    )
    tags = ak.layout.Index8(np.array([1, 1, 0, 0, 1, 0, 1, 1], dtype=np.int8))
    index = ak.layout.Index32(np.array([0, 1, 0, 1, 2, 2, 4, 3], dtype=np.int32))
    array = ak.layout.UnionArray8_32(tags, index, [content0, content1])
    array = v1_to_v2(array)

    with pytest.raises(ValueError):
        array.sort(axis=1, ascending=True, stable=False)


def test_sort_strings():
    content = ak.from_iter(["one", "two", "three", "four", "five"], highlevel=False)
    assert ak.to_list(content) == ["one", "two", "three", "four", "five"]
    content = v1_to_v2(content)

    assert ak.to_list(content.sort(axis=0, ascending=True, stable=False)) == [
        "five",
        "four",
        "one",
        "three",
        "two",
    ]
    assert ak.to_list(content.sort(axis=0, ascending=False, stable=False)) == [
        "two",
        "three",
        "one",
        "four",
        "five",
    ]


def test_sort_bytestrings():
    array = ak.from_iter(
        [b"one", b"two", b"three", b"two", b"two", b"one", b"three"], highlevel=False
    )
    array = v1_to_v2(array)
    assert ak.to_list(array) == [
        b"one",
        b"two",
        b"three",
        b"two",
        b"two",
        b"one",
        b"three",
    ]

    assert ak.to_list(array.sort(axis=0, ascending=True, stable=False)) == [
        b"one",
        b"one",
        b"three",
        b"three",
        b"two",
        b"two",
        b"two",
    ]

    assert ak.to_list(array.argsort(axis=0, ascending=True, stable=True)) == [
        0,
        5,
        2,
        6,
        1,
        3,
        4,
    ]


def test_sort_zero_length_arrays():
    array = ak.layout.IndexedArray64(
        ak.layout.Index64([]), ak.layout.NumpyArray([1, 2, 3])
    )
    array = v1_to_v2(array)
    assert ak.to_list(array) == []
    assert ak.to_list(array.sort()) == []
    assert ak.to_list(array.argsort()) == []

    content = ak.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    mask = ak.layout.Index8([])
    array = ak.layout.ByteMaskedArray(mask, content, valid_when=False)
    array = v1_to_v2(array)
    assert ak.to_list(array) == []
    assert ak.to_list(array.sort()) == []
    assert ak.to_list(array.argsort()) == []

    array = ak.layout.NumpyArray([])
    array = v1_to_v2(array)
    assert ak.to_list(array) == []
    assert ak.to_list(array.sort()) == []
    assert ak.to_list(array.argsort()) == []

    array = ak.layout.RecordArray([])
    array = v1_to_v2(array)
    assert ak.to_list(array) == []
    assert ak.to_list(array.sort()) == []
    assert ak.to_list(array.argsort()) == []

    content = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts1 = ak.layout.Index64([])
    stops1 = ak.layout.Index64([])
    offsets1 = ak.layout.Index64(np.array([0]))
    array = ak.layout.ListArray64(starts1, stops1, content)
    array = v1_to_v2(array)
    assert ak.to_list(array) == []
    assert ak.to_list(array.sort()) == []
    assert ak.to_list(array.argsort()) == []

    array = ak.layout.ListOffsetArray64(offsets1, content)
    array = v1_to_v2(array)
    assert ak.to_list(array) == []
    assert ak.to_list(array.sort()) == []
    assert ak.to_list(array.argsort()) == []


@pytest.mark.skip(reason="FIXME: cannot sort an irreducible UnionArray")
def test_UnionArray_FIXME():
    content0 = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    content1 = ak.from_iter(["one", "two", "three", "four", "five"], highlevel=False)
    tags = ak.layout.Index8([])
    index = ak.layout.Index32([])
    array = ak.layout.UnionArray8_32(tags, index, [content0, content1])
    array = v1_to_v2(array)
    assert ak.to_list(array) == []

    assert ak.to_list(array.sort()) == []
    assert ak.to_list(array.argsort()) == []



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


array = ak.Array([[None, 2], None, [4, None]])


def test_fill_none_axis_0():
    filled = ak.fill_none(array, 10, axis=0)
    assert ak.to_list(filled) == [[None, 2], 10, [4, None]]


def test_fill_none_axis_1():
    filled = ak.fill_none(array, 10, axis=1)
    assert ak.to_list(filled) == [[10, 2], None, [4, 10]]


def test_fill_none_axis_none():
    filled = ak.fill_none(array, 10, axis=None)
    assert ak.to_list(filled) == [[10, 2], 10, [4, 10]]


def test_fill_none_axis_deprecated():
    ak.deprecations_as_errors = False

    with pytest.deprecated_call():
        filled = ak.fill_none(array, 10)
        assert ak.to_list(filled) == [[None, 2], 10, [4, None]]

        with pytest.deprecated_call():
            filled_twice = ak.fill_none(filled, 20)
            assert ak.to_list(filled_twice) == [[20, 2], 10, [4, 20]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_basic():
    content = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4]))

    ind = np.array([2, 2, 0, 3, 4], dtype=np.int32)
    index = ak.layout.Index32(ind)
    array = ak.layout.IndexedArray32(index, content)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = np.array([2, 2, 0, 3, 4], dtype=np.uint32)
    index = ak.layout.IndexU32(ind)
    array = ak.layout.IndexedArrayU32(index, content)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = np.array([2, 2, 0, 3, 4], dtype=np.int64)
    index = ak.layout.Index64(ind)
    array = ak.layout.IndexedArray64(index, content)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = np.array([2, 2, 0, 3, 4], dtype=np.int32)
    index = ak.layout.Index32(ind)
    array = ak.layout.IndexedOptionArray32(index, content)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = np.array([2, 2, 0, 3, 4], dtype=np.int64)
    index = ak.layout.Index64(ind)
    array = ak.layout.IndexedOptionArray64(index, content)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 1.1, 4.4]


def test_type():
    content = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    index = ak.layout.Index32(np.array([2, 2, 0, 3, 4], dtype=np.int32))
    array = ak.layout.IndexedArray32(index, content)
    assert ak.type(array) == ak.types.PrimitiveType("float64")
    array = ak.layout.IndexedOptionArray32(index, content)
    assert ak.type(array) == ak.types.OptionType(ak.types.PrimitiveType("float64"))


def test_null():
    content = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    index = ak.layout.Index64(np.array([2, 2, 0, -1, 4], dtype=np.int64))
    array = ak.layout.IndexedOptionArray64(index, content)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, None, 4.4]


def test_carry():
    content = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    index = ak.layout.Index64(np.array([2, 2, 0, 3, 4], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray64(index, content)
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, indexedarray)
    assert ak.to_list(listoffsetarray) == [[2.2, 2.2, 0.0], [], [3.3, 4.4]]
    assert ak.to_list(listoffsetarray[::-1]) == [[3.3, 4.4], [], [2.2, 2.2, 0.0]]
    assert ak.to_list(listoffsetarray[[2, 0]]) == [[3.3, 4.4], [2.2, 2.2, 0.0]]
    assert ak.to_list(listoffsetarray[[2, 0], 1]) == [4.4, 2.2]  # invokes carry
    assert ak.to_list(listoffsetarray[2:, 1]) == [4.4]  # invokes carry

    index = ak.layout.Index64(np.array([2, 2, 0, 3, -1], dtype=np.int64))
    indexedarray = ak.layout.IndexedOptionArray64(index, content)
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, indexedarray)
    assert ak.to_list(listoffsetarray) == [[2.2, 2.2, 0.0], [], [3.3, None]]
    assert ak.to_list(listoffsetarray[::-1]) == [[3.3, None], [], [2.2, 2.2, 0.0]]
    assert ak.to_list(listoffsetarray[[2, 0]]) == [[3.3, None], [2.2, 2.2, 0.0]]
    assert ak.to_list(listoffsetarray[[2, 0], 1]) == [None, 2.2]  # invokes carry
    assert ak.to_list(listoffsetarray[2:, 1]) == [None]  # invokes carry


def test_others():
    content = ak.layout.NumpyArray(
        np.array(
            [[0.0, 0.0], [0.1, 1.0], [0.2, 2.0], [0.3, 3.0], [0.4, 4.0], [0.5, 5.0]]
        )
    )
    index = ak.layout.Index64(np.array([4, 0, 3, 1, 3], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray64(index, content)
    assert indexedarray[3, 0] == 0.1
    assert indexedarray[3, 1] == 1.0
    assert ak.to_list(indexedarray[3, ::-1]) == [1.0, 0.1]
    assert ak.to_list(indexedarray[3, [1, 1, 0]]) == [1.0, 1.0, 0.1]
    assert ak.to_list(indexedarray[3:, 0]) == [0.1, 0.3]
    assert ak.to_list(indexedarray[3:, 1]) == [1.0, 3.0]
    assert ak.to_list(indexedarray[3:, ::-1]) == [[1.0, 0.1], [3.0, 0.3]]
    assert ak.to_list(indexedarray[3:, [1, 1, 0]]) == [[1.0, 1.0, 0.1], [3.0, 3.0, 0.3]]


def test_missing():
    content = ak.layout.NumpyArray(
        np.array(
            [[0.0, 0.0], [0.1, 1.0], [0.2, 2.0], [0.3, 3.0], [0.4, 4.0], [0.5, 5.0]]
        )
    )
    index = ak.layout.Index64(np.array([4, 0, 3, -1, 3], dtype=np.int64))
    indexedarray = ak.layout.IndexedOptionArray64(index, content)
    assert ak.to_list(indexedarray[3:, 0]) == [None, 0.3]
    assert ak.to_list(indexedarray[3:, 1]) == [None, 3.0]
    assert ak.to_list(indexedarray[3:, ::-1]) == [None, [3.0, 0.3]]
    assert ak.to_list(indexedarray[3:, [1, 1, 0]]) == [None, [3.0, 3.0, 0.3]]


def test_highlevel():
    content = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    index = ak.layout.Index64(np.array([2, 2, 0, -1, 4], dtype=np.int64))
    array = ak.Array(ak.layout.IndexedOptionArray64(index, content), check_valid=True)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, None, 4.4]
    assert str(array) == "[2.2, 2.2, 0, None, 4.4]"
    assert repr(array) == "<Array [2.2, 2.2, 0, None, 4.4] type='5 * ?float64'>"


def test_builder():
    assert ak.to_list(ak.Array([1.1, 2.2, 3.3, None, 4.4], check_valid=True)) == [
        1.1,
        2.2,
        3.3,
        None,
        4.4,
    ]
    assert ak.to_list(ak.Array([None, 2.2, 3.3, None, 4.4], check_valid=True)) == [
        None,
        2.2,
        3.3,
        None,
        4.4,
    ]

    assert ak.to_list(
        ak.Array([[1.1, 2.2, 3.3], [], [None, 4.4]], check_valid=True)
    ) == [[1.1, 2.2, 3.3], [], [None, 4.4]]
    assert ak.to_list(
        ak.Array([[1.1, 2.2, 3.3], [], None, [None, 4.4]], check_valid=True)
    ) == [[1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert ak.to_list(
        ak.Array([[1.1, 2.2, 3.3], None, [], [None, 4.4]], check_valid=True)
    ) == [[1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert ak.to_list(
        ak.Array([[1.1, 2.2, 3.3], None, [], [None, 4.4]], check_valid=True)
    ) != [[1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert ak.to_list(
        ak.Array([[None, 1.1, 2.2, 3.3], [], [None, 4.4]], check_valid=True)
    ) == [[None, 1.1, 2.2, 3.3], [], [None, 4.4]]
    assert ak.to_list(
        ak.Array([[None, 1.1, 2.2, 3.3], [], None, [None, 4.4]], check_valid=True)
    ) == [[None, 1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert ak.to_list(
        ak.Array([[None, 1.1, 2.2, 3.3], None, [], [None, 4.4]], check_valid=True)
    ) == [[None, 1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert ak.to_list(
        ak.Array([[None, 1.1, 2.2, 3.3], None, [], [None, 4.4]], check_valid=True)
    ) != [[None, 1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert ak.to_list(
        ak.Array([None, [1.1, 2.2, 3.3], [], [None, 4.4]], check_valid=True)
    ) == [None, [1.1, 2.2, 3.3], [], [None, 4.4]]
    assert ak.to_list(
        ak.Array([None, [1.1, 2.2, 3.3], [], None, [None, 4.4]], check_valid=True)
    ) == [None, [1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert ak.to_list(
        ak.Array([None, [1.1, 2.2, 3.3], None, [], [None, 4.4]], check_valid=True)
    ) == [None, [1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert ak.to_list(
        ak.Array([None, [1.1, 2.2, 3.3], None, [], [None, 4.4]], check_valid=True)
    ) != [None, [1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert ak.to_list(ak.Array([None, None, None, None, None], check_valid=True)) == [
        None,
        None,
        None,
        None,
        None,
    ]
    assert ak.to_list(
        ak.Array([[None, None, None], [], [None, None]], check_valid=True)
    ) == [[None, None, None], [], [None, None]]


def test_json():
    assert ak.to_list(ak.Array("[1.1, 2.2, 3.3, null, 4.4]", check_valid=True)) == [
        1.1,
        2.2,
        3.3,
        None,
        4.4,
    ]
    assert ak.to_list(ak.Array("[null, 2.2, 3.3, null, 4.4]", check_valid=True)) == [
        None,
        2.2,
        3.3,
        None,
        4.4,
    ]

    assert ak.to_list(
        ak.Array("[[1.1, 2.2, 3.3], [], [null, 4.4]]", check_valid=True)
    ) == [[1.1, 2.2, 3.3], [], [None, 4.4]]
    assert ak.to_list(
        ak.Array("[[1.1, 2.2, 3.3], [], null, [null, 4.4]]", check_valid=True)
    ) == [[1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert ak.to_list(
        ak.Array("[[1.1, 2.2, 3.3], null, [], [null, 4.4]]", check_valid=True)
    ) == [[1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert ak.to_list(
        ak.Array("[[1.1, 2.2, 3.3], null, [], [null, 4.4]]", check_valid=True)
    ) != [[1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert ak.to_list(
        ak.Array("[[null, 1.1, 2.2, 3.3], [], [null, 4.4]]", check_valid=True)
    ) == [[None, 1.1, 2.2, 3.3], [], [None, 4.4]]
    assert ak.to_list(
        ak.Array("[[null, 1.1, 2.2, 3.3], [], null, [null, 4.4]]", check_valid=True)
    ) == [[None, 1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert ak.to_list(
        ak.Array("[[null, 1.1, 2.2, 3.3], null, [], [null, 4.4]]", check_valid=True)
    ) == [[None, 1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert ak.to_list(
        ak.Array("[[null, 1.1, 2.2, 3.3], null, [], [null, 4.4]]", check_valid=True)
    ) != [[None, 1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert ak.to_list(
        ak.Array("[null, [1.1, 2.2, 3.3], [], [null, 4.4]]", check_valid=True)
    ) == [None, [1.1, 2.2, 3.3], [], [None, 4.4]]
    assert ak.to_list(
        ak.Array("[null, [1.1, 2.2, 3.3], [], null, [null, 4.4]]", check_valid=True)
    ) == [None, [1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert ak.to_list(
        ak.Array("[null, [1.1, 2.2, 3.3], null, [], [null, 4.4]]", check_valid=True)
    ) == [None, [1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert ak.to_list(
        ak.Array("[null, [1.1, 2.2, 3.3], null, [], [null, 4.4]]", check_valid=True)
    ) != [None, [1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert ak.to_list(ak.Array("[null, null, null, null, null]", check_valid=True)) == [
        None,
        None,
        None,
        None,
        None,
    ]
    assert ak.to_list(
        ak.Array("[[null, null, null], [], [null, null]]", check_valid=True)
    ) == [[None, None, None], [], [None, None]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

from awkward._v2.types.type import Type  # noqa: F401
from awkward._v2.types.unknowntype import UnknownType  # noqa: F401
from awkward._v2.types.numpytype import NumpyType  # noqa: F401
from awkward._v2.types.regulartype import RegularType  # noqa: F401
from awkward._v2.types.listtype import ListType  # noqa: F401
from awkward._v2.types.recordtype import RecordType  # noqa: F401
from awkward._v2.types.optiontype import OptionType  # noqa: F401
from awkward._v2.types.uniontype import UnionType  # noqa: F401
from awkward._v2.types.arraytype import ArrayType  # noqa: F401



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_parameters_on_arrays():
    a = ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    assert a.parameters == {}
    a.setparameter("one", ["two", 3, {"four": 5}])
    assert a.parameters == {"one": ["two", 3, {"four": 5}]}

    b = ak.layout.ListOffsetArray64(
        ak.layout.Index64(np.array([0, 3, 3, 5], dtype=np.int64)), a
    )
    assert b.parameters == {}
    assert b.content.parameters == {"one": ["two", 3, {"four": 5}]}
    b.setparameter("what", "ever")
    assert b.parameters == {"what": "ever"}
    assert b.content.parameters == {"one": ["two", 3, {"four": 5}]}


def test_string2():
    content = ak.layout.NumpyArray(
        np.array([ord(x) for x in "heythere"], dtype=np.uint8),
        parameters={"__array__": "char", "encoding": "utf-8"},
    )
    listoffsetarray = ak.layout.ListOffsetArray64(
        ak.layout.Index64(np.array([0, 3, 3, 8])),
        content,
        parameters={"__array__": "string"},
    )
    a = ak.Array(listoffsetarray, check_valid=True)

    assert isinstance(a, ak.Array)
    assert ak.to_list(a) == ["hey", "", "there"]

    if ak._util.py27:
        assert str(a) == "[u'hey', u'', u'there']"
        assert repr(a[0]) == "u'hey'"
        assert repr(a[1]) == "u''"
        assert repr(a[2]) == "u'there'"
    else:
        assert str(a) == "['hey', '', 'there']"
        assert repr(a[0]) == "'hey'"
        assert repr(a[1]) == "''"
        assert repr(a[2]) == "'there'"


def test_dress():
    class Dummy(ak.highlevel.Array):
        def __repr__(self):
            return "<Dummy {0}>".format(str(self))

    ns = {"Dummy": Dummy}

    x = ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    x.setparameter("__array__", "Dummy")
    a = ak.Array(x, behavior=ns, check_valid=True)
    assert repr(a) == "<Dummy [1.1, 2.2, 3.3, 4.4, 5.5]>"

    x2 = ak.layout.ListOffsetArray64(
        ak.layout.Index64(np.array([0, 3, 3, 5], dtype=np.int64)),
        ak.layout.NumpyArray(
            np.array([1.1, 2.2, 3.3, 4.4, 5.5]), parameters={"__array__": "Dummy"}
        ),
    )
    a2 = ak.Array(x2, behavior=ns, check_valid=True)
    assert (
        repr(a2)
        == "<Array [<Dummy [1.1, 2.2, 3.3]>, ... ] type='3 * var * float64[parameters={\"__ar...'>"
    )
    assert repr(a2[0]) == "<Dummy [1.1, 2.2, 3.3]>"
    assert repr(a2[1]) == "<Dummy []>"
    assert repr(a2[2]) == "<Dummy [4.4, 5.5]>"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_integerindex_null():
    a = ak.Array([[0, 1, 2], None, [5, 6], [7]])
    b = ak.Array([[0, 1, 2], [3, 4], [5, 6], [7]])
    c = ak.Array([[1], [1], [0], [0]])
    d = ak.Array([[1], None, [0], [0]])
    e = ak.Array([[1], None, None, [0]])

    a = v1_to_v2(a.layout)
    b = v1_to_v2(b.layout)
    c = v1_to_v2(c.layout)
    d = v1_to_v2(d.layout)
    e = v1_to_v2(e.layout)

    assert ak.to_list(a[c]) == [[1], None, [5], [7]]
    assert ak.to_list(a[d]) == [[1], None, [5], [7]]
    assert ak.to_list(a[e]) == [[1], None, None, [7]]
    assert ak.to_list(b[c]) == [[1], [4], [5], [7]]
    assert ak.to_list(b[d]) == [[1], None, [5], [7]]
    assert ak.to_list(b[e]) == [[1], None, None, [7]]

    assert a.typetracer[c].form == a[c].form
    assert a.typetracer[d].form == a[d].form
    assert a.typetracer[e].form == a[e].form
    assert b.typetracer[c].form == b[c].form
    assert b.typetracer[d].form == b[d].form
    assert b.typetracer[e].form == b[e].form


def test_boolindex_null():
    a = ak.Array([[0, 1, 2], None, [5, 6]])
    b = ak.Array([[0, 1, 2], [3, 4], [5, 6]])
    c = ak.Array([[False, True, False], [False, True], [True, False]])
    d = ak.Array([[False, True, False], None, [True, False]])
    e = ak.Array([[False, True, False], None, None])
    b2 = b.mask[[True, False, True]]

    a = v1_to_v2(a.layout)
    b = v1_to_v2(b.layout)
    c = v1_to_v2(c.layout)
    d = v1_to_v2(d.layout)
    e = v1_to_v2(e.layout)

    assert ak.to_list(a[c]) == [[1], None, [5]]
    assert ak.to_list(a[d]) == [[1], None, [5]]
    assert ak.to_list(a[e]) == [[1], None, None]
    assert ak.to_list(b[c]) == [[1], [4], [5]]
    assert ak.to_list(b[d]) == [[1], None, [5]]
    assert ak.to_list(b[e]) == [[1], None, None]

    assert a.typetracer[c].form == a[c].form
    assert a.typetracer[d].form == a[d].form
    assert a.typetracer[e].form == a[e].form
    assert b.typetracer[c].form == b[c].form
    assert b.typetracer[d].form == b[d].form
    assert b.typetracer[e].form == b[e].form

    b2 = v1_to_v2(b2.layout)

    assert ak.to_list(b2[c]) == [[1], None, [5]]
    assert ak.to_list(b2[d]) == [[1], None, [5]]
    assert ak.to_list(b2[e]) == [[1], None, None]

    assert b2.typetracer[c].form == b2[c].form
    assert b2.typetracer[d].form == b2[d].form
    assert b2.typetracer[e].form == b2[e].form


def test_integerindex_null_more():
    f = ak.Array([[0, None, 2], None, [3, 4], []])
    g1 = ak.Array([[1, 2, None], None, [], [None]])
    g2 = ak.Array([[], None, None, []])
    g3 = ak.Array([[], [], [], []])

    f = v1_to_v2(f.layout)
    g1 = v1_to_v2(g1.layout)
    g2 = v1_to_v2(g2.layout)
    g3 = v1_to_v2(g3.layout)

    assert ak.to_list(f[g1]) == [[None, 2, None], None, [], [None]]
    assert ak.to_list(f[g2]) == [[], None, None, []]
    assert ak.to_list(f[g3]) == [[], None, [], []]
    assert f.typetracer[g1].form == f[g1].form
    assert f.typetracer[g2].form == f[g2].form
    assert f.typetracer[g3].form == f[g3].form

    a = ak.Array([[0, 1, 2, None], None])
    b = ak.Array([[2, 1, None, 3], None])

    a = v1_to_v2(a.layout)
    b = v1_to_v2(b.layout)

    assert ak.to_list(a[b]) == [[2, 1, None, None], None]
    assert a.typetracer[b].form == a[b].form

    b = ak.Array([[2, 1, None, 3], []])
    b = v1_to_v2(b.layout)

    assert ak.to_list(a[b]) == [[2, 1, None, None], None]
    assert a.typetracer[b].form == a[b].form

    b = ak.Array([[2, 1, None, 3], [0, 1]])
    b = v1_to_v2(b.layout)
    assert ak.to_list(a[b]) == [[2, 1, None, None], None]
    assert a.typetracer[b].form == a[b].form


@pytest.mark.skip(reason="FIXME: erroneous slicing case")
def test_integerindex_null_more_2():
    a = ak.Array([[[0, 1, 2, None], None], [[3, 4], [5]], None, [[6]]])
    b = ak.Array([[[2, 1, None, 3], [0, 1]], [[0], None], None, [None]])
    c = ak.Array(
        [
            [[False, True, None, False], [False, True]],
            [[True, False], None],
            None,
            [None],
        ]
    )

    a = v1_to_v2(a.layout)
    b = v1_to_v2(b.layout)
    c = v1_to_v2(c.layout)

    assert ak.to_list(a[b]) == [
        [[2, 1, None, None], None],
        [[3], None],
        None,
        [None],
    ]
    assert ak.to_list(a[c]) == [[[1, None], None], [[4], None], None, [None]]
    assert a.typetracer[c].form == a[c].form


def test_silly_stuff():
    a = ak.Array([[0, 1, 2], 3])
    b = [[2], [0]]

    a = v1_to_v2(a.layout)

    with pytest.raises(IndexError):
        a[b]

    a = ak.Array([[0, 1, 2], [3, 4], [5, 6], [7]])
    a = v1_to_v2(a.layout)

    b = ak.Array([[0, 2], None, [1], None])
    b = v1_to_v2(b.layout)

    assert ak.to_list(a[b]) == [[0, 2], None, [6], None]

    b = ak.Array([[0, 2], None, None, None, None, None])
    b = v1_to_v2(b.layout)

    with pytest.raises(IndexError):
        a[b]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array(["123", "45", "00001", "    0000   ", "-2.6", "-3", "1e1", "1E1"])
    assert ak.strings_astype(array, np.float64).tolist() == [
        123,
        45,
        1,
        0,
        -2.6,
        -3,
        10,
        10,
    ]
    assert ak.strings_astype(
        array[[True, True, True, True, False, True, False, False]], np.int64
    ).tolist() == [123, 45, 1, 0, -3]

    array = ak.Array(
        [["123", "45", "00001"], [], ["    0000   ", "-2.6"], ["-3"], ["1e1", "1E1"]]
    )
    assert ak.strings_astype(array, np.float64).tolist() == [
        [123, 45, 1],
        [],
        [0, -2.6],
        [-3],
        [10, 10],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    layout = ak.layout.IndexedArray64(
        ak.layout.Index64(np.array([3, 1, 0, 2])),
        ak.layout.ListOffsetArray64(
            ak.layout.Index64(np.array([0, 3, 6, 9, 12])),
            ak.layout.NumpyArray(np.array([0, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 3])),
        ),
    )

    assert ak.unflatten(
        layout, ak.flatten(ak.run_lengths(layout)), axis=1
    ).tolist() == [[[3, 3, 3]], [[1, 1, 1]], [[0, 0, 0]], [[2, 2], [3]]]
